# File ported from: https://raw.githubusercontent.com/deepmind/dm-haiku/master/haiku/_src/embed.py

"""Modules for performing embedding lookups in Haiku."""

import enum
from typing import Optional, Union

import jax
import jax.numpy as jnp

from elegy.module import Module
from elegy import initializers, types


class EmbedLookupStyle(enum.Enum):
    """How to return the embedding matrices given IDs."""

    ARRAY_INDEX = 1
    ONE_HOT = 2


# Needed for members to show in sphinx docs.
for style in EmbedLookupStyle:
    style.__doc__ = style.name


class Embedding(Module):
    """Module for embedding tokens in a low-dimensional space."""

    def __init__(
        self,
        vocab_size: Optional[int] = None,
        embed_dim: Optional[int] = None,
        embedding_matrix: Optional[jnp.ndarray] = None,
        w_init: Optional[types.Initializer] = None,
        lookup_style: Union[str, EmbedLookupStyle] = "ARRAY_INDEX",
        name: Optional[str] = None,
    ):
        """
        Constructs an Embed module.

        Args:
            vocab_size: The number of unique tokens to embed. If not provided, an
                existing vocabulary matrix from which ``vocab_size`` can be inferred
                must be provided as ``existing_vocab``.
            embed_dim: Number of dimensions to assign to each embedding. If an
                existing vocabulary matrix initializes the module, this should not be
                provided as it will be inferred.
            embedding_matrix: A matrix-like object equivalent in size to
                ``[vocab_size, embed_dim]``. If given, it is used as the initial value
                for the embedding matrix and neither ``vocab_size`` or ``embed_dim``
                need be given. If they are given, their values are checked to be
                consistent with the dimensions of ``embedding_matrix``.
            w_init: An initializer for the embeddings matrix. As a default,
                embeddings are initialized via a truncated normal distribution.
            lookup_style: One of the enum values of :class:`EmbedLookupStyle`
                determining how to access the value of the embeddings given an ID.
                Regardless the input should be a dense array of integer values
                representing ids. This setting changes how internally this module maps
                those ides to embeddings. The result is the same, but the speed and
                memory tradeoffs are different. It default to using numpy-style array
                indexing. This value is only the default for the module, and at any
                given invocation can be overridden in :meth:`__call__`.
            name: Optional name for this module.

        Raises:
            ValueError: If none of ``embed_dim``, ``embedding_matrix`` and
                ``vocab_size`` are supplied, or if ``embedding_matrix`` is supplied
                and ``embed_dim`` or ``vocab_size`` is not consistent with the
                supplied matrix.
        """
        super().__init__(name=name)
        if embedding_matrix is None and not (vocab_size and embed_dim):
            raise ValueError(
                "Embedding must be supplied either with an initial `embedding_matrix` "
                "or with `embed_dim` and `vocab_size`."
            )
        if embedding_matrix is not None:
            embedding_matrix = jnp.asarray(embedding_matrix)
            if vocab_size and embedding_matrix.shape[0] != vocab_size:
                raise ValueError(
                    "An `embedding_matrix` was supplied but the `vocab_size` of "
                    f"{vocab_size} was not consistent with its shape "
                    f"{embedding_matrix.shape}."
                )
            if embed_dim and embedding_matrix.shape[1] != embed_dim:
                raise ValueError(
                    "An `embedding_matrix` was supplied but the `embed_dim` of "
                    f"{embed_dim} was not consistent with its shape "
                    f"{embedding_matrix.shape}."
                )
            self.embeddings = self.add_parameter(
                "embeddings",
                embedding_matrix.shape,
                initializer=lambda _, __: embedding_matrix,
            )
        else:
            assert embed_dim is not None
            assert vocab_size is not None

            w_init = w_init or initializers.TruncatedNormal()
            self.embeddings = self.add_parameter(
                "embeddings", [vocab_size, embed_dim], initializer=w_init
            )

        self.vocab_size = vocab_size or embedding_matrix.shape[0]
        self.embed_dim = embed_dim or embedding_matrix.shape[1]
        self.lookup_style = lookup_style

    def call(
        self,
        ids: jnp.ndarray,
        lookup_style: Optional[Union[str, EmbedLookupStyle]] = None,
    ) -> jnp.ndarray:
        r"""
        Lookup embeddings.

        Looks up an embedding vector for each value in ``ids``. All ids must be
        within ``[0, vocab_size)`` to prevent ``NaN``\ s from propagating.

        Args:
            ids: integer array.
            lookup_style: Overrides the ``lookup_style`` given in the constructor.

        Returns:
            Tensor of ``ids.shape + [embedding_dim]``.

        Raises:
            AttributeError: If ``lookup_style`` is not valid.
            ValueError: If ``ids`` is not an integer array.
        """
        # TODO(tomhennigan) Consider removing asarray here.
        ids = jnp.asarray(ids)
        if not jnp.issubdtype(ids.dtype, jnp.integer):
            raise ValueError(
                "Embedding's __call__ method must take an array of "
                "integer dtype but was called with an array of "
                f"{ids.dtype}"
            )

        lookup_style = lookup_style or self.lookup_style
        if isinstance(lookup_style, str):
            lookup_style = getattr(EmbedLookupStyle, lookup_style.upper())

        if lookup_style == EmbedLookupStyle.ARRAY_INDEX:
            return self.embeddings[(ids,)]

        elif lookup_style == EmbedLookupStyle.ONE_HOT:
            one_hot_ids = jax.nn.one_hot(ids, self.vocab_size)[..., None]
            return (self.embeddings * one_hot_ids).sum(axis=-2)

        else:
            raise NotImplementedError(f"{lookup_style} is not supported by Embedding.")



from contextlib import contextmanager
from elegy import utils, types
import functools
import threading
import typing as tp
from dataclasses import dataclass
import jax

import jax.numpy as jnp
import numpy as np

from elegy import types

__all__ = [
    "add_loss",
    "add_metric",
    "add_summary",
    "get_losses",
    "losses_active",
    "get_metrics",
    "metrics_active",
    "get_summaries",
    "summaries_active",
    "context",
    "jit",
]


class HooksContext(types.Protocol):
    losses: tp.Optional[types.Logs]
    metrics: tp.Optional[types.Logs]
    summaries: tp.Optional[types.Summaries]


@dataclass
class _HooksContext(threading.local):
    losses: tp.Optional[types.Logs]
    metrics: tp.Optional[types.Logs]
    summaries: tp.Optional[types.Summaries]


LOCAL: HooksContext = _HooksContext(
    losses=None,
    metrics=None,
    summaries=None,
)


# ----------------------------------------------------------------
# hooks
# ----------------------------------------------------------------


def add_loss(name: str, value: types.Scalar) -> None:
    """
    A hook that lets you define a loss within a [`module`][elegy.module.Module].

    ```python
    w = self.add_parameter("w", [3, 5], initializer=jnp.ones)

    # L2 regularization penalty
    elegy.hooks.add_loss("l2_regularization", 0.01 * jnp.mean(w ** 2))
    ```

    Arguments:
        name: The name of the loss. If a `name` is repeated on
            different calls values will be added together.
        value: The value for the loss.
    """
    if LOCAL.losses is None:
        return

    if not name.endswith("loss"):
        name += "_loss"

    if name in LOCAL.losses:
        LOCAL.losses[name] = LOCAL.losses[name] + value
    else:
        LOCAL.losses[name] = value


def add_metric(name: str, value: types.Scalar) -> None:
    """
    A hook that lets you define a metric within a [`module`][elegy.module.Module].

    ```python
    y = jax.nn.relu(x)
    elegy.hooks.add_metric("activation_mean", jnp.mean(y))
    ```

    Arguments:
        name: The name of the loss. If a metric with the same
            `name` already exists a unique identifier will be generated.
        value: The value for the metric.
    """
    if LOCAL.metrics is None:
        return

    # name = f"{base_name()}/{name}"
    name = utils.get_unique_name(set(LOCAL.metrics), name)
    LOCAL.metrics[name] = value


def add_summary(
    path: types.Path,
    module: tp.Any,
    value: tp.Any,
) -> None:
    """
    A hook that lets you define a summary in the current module. Its primary
    use is to keep track of certain values as they flow through the network
    so [`Model.summary`][elegy.model.model.Model.summary] can show a representation of architecture.

    ```python
    def call(self, x):
        ...
        y = jax.nn.relu(x)
        elegy.hooks.add_summary("relu", y)
        ...
    ```

    Arguments:
        module_or_name: The name of the summary or alternatively the module that this summary will represent.
            If a summary with the same name already exists a unique identifier will be generated.
        value: The value for the summary.
    """

    if not summaries_active():
        return

    LOCAL.summaries.append(types.Summary(path, module, value))


def get_losses() -> types.Logs:
    if LOCAL.losses is None:
        return {}

    return LOCAL.losses.copy()


def losses_active() -> bool:
    return LOCAL.losses is not None


def get_metrics() -> types.Logs:
    if LOCAL.metrics is None:
        return {}

    return LOCAL.metrics.copy()


def metrics_active() -> bool:
    return LOCAL.metrics is not None


def get_summaries() -> types.Summaries:
    if LOCAL.summaries is None:
        return []

    return LOCAL.summaries.copy()


def summaries_active() -> bool:
    return LOCAL.summaries is not None


# ----------------------------------------------------------------
# contexts
# ----------------------------------------------------------------


def context(
    *,
    losses: tp.Union[types.Logs, bool, None] = None,
    metrics: tp.Union[types.Logs, bool, None] = None,
    summaries: tp.Union[types.Summaries, bool, None] = None,
    set_all: bool = False,
) -> tp.ContextManager[None]:

    if set_all:
        if losses is None:
            losses = True
        if metrics is None:
            metrics = True
        if summaries is None:
            summaries = True

    if isinstance(losses, bool):
        losses = {} if losses else None

    if isinstance(metrics, bool):
        metrics = {} if metrics else None

    if isinstance(summaries, bool):
        summaries = [] if summaries else None

    return _context(
        losses=losses,
        metrics=metrics,
        summaries=summaries,
    )


@contextmanager
def _context(
    losses: tp.Optional[types.Logs],
    metrics: tp.Optional[types.Logs],
    summaries: tp.Optional[types.Summaries],
) -> tp.Iterator[None]:

    prev_losses = LOCAL.losses
    prev_metrics = LOCAL.metrics
    prev_summaries = LOCAL.summaries

    LOCAL.losses = losses
    LOCAL.metrics = metrics
    LOCAL.summaries = summaries

    try:
        yield
    finally:
        LOCAL.losses = prev_losses
        LOCAL.metrics = prev_metrics
        LOCAL.summaries = prev_summaries


# -------------------------------------------------------------
# transforms
# -------------------------------------------------------------


class TransformtOutput(tp.NamedTuple):
    output: tp.Any
    losses: tp.Optional[types.Logs]
    metrics: tp.Optional[types.Logs]
    summaries: tp.Optional[tp.List[tp.Any]]


class DynamicArgs(tp.NamedTuple):
    losses: tp.Optional[types.Logs]
    metrics: tp.Optional[types.Logs]
    summaries: tp.Optional[tp.List[tp.Any]]


class StaticArgs(tp.NamedTuple):
    pass


def _update_local_context(
    losses: tp.Optional[types.Logs],
    metrics: tp.Optional[types.Logs],
    summaries: tp.Optional[types.Summaries],
):
    if LOCAL.losses is not None and losses is not None:
        LOCAL.losses.clear()
        LOCAL.losses.update(losses)

    if LOCAL.metrics is not None and metrics is not None:
        LOCAL.metrics.clear()
        LOCAL.metrics.update(metrics)

    if LOCAL.summaries is not None and summaries is not None:
        LOCAL.summaries.clear()
        LOCAL.summaries.extend(summaries)


def jit(
    f,
    **kwargs,
):
    def _transform_fn(
        *args,
    ) -> TransformtOutput:

        # extract input context
        dynamic_args: DynamicArgs
        static_args: StaticArgs

        # static_args is unused because they dont need to be set back
        static_args, dynamic_args = args[:2]  # get from beginning
        args = args[2:]

        (losses, metrics, summaries) = dynamic_args

        # perform updates
        _update_local_context(losses, metrics, summaries)

        # call
        output = f(*args)

        # add outputs context
        return TransformtOutput(
            output=output,
            losses=LOCAL.losses,
            metrics=LOCAL.metrics,
            summaries=LOCAL.summaries,
        )

    # transform kwargs
    static_argnums = kwargs.pop("static_argnums", [])
    if isinstance(static_argnums, int):
        static_argnums = [static_argnums]
    static_argnums = [0] + [i + 2 for i in static_argnums]

    # make fn
    transform_fn = jax.jit(
        _transform_fn,
        static_argnums=static_argnums,
        **kwargs,
    )

    @functools.wraps(f)
    def wrapper(*args):

        # add input context
        dynamic_args = DynamicArgs(
            losses=LOCAL.losses,
            metrics=LOCAL.metrics,
            summaries=LOCAL.summaries,
        )
        static_args = StaticArgs()
        # put them first because of static_args
        args = (static_args, dynamic_args) + args

        # call and patch
        (
            output,
            losses,
            metrics,
            summaries,
        ) = transform_fn(*args)

        # perform updates
        _update_local_context(losses, metrics, summaries)

        return output

    return wrapper


def hooks_aware(transform):
    def transform_wrapper(
        f,
        **kwargs,
    ):
        def f_wrapper(
            *args,
        ) -> TransformtOutput:

            # extract input context
            dynamic_args: DynamicArgs

            # static_args is unused because they dont need to be set back
            dynamic_args = args[0]  # get from beginning
            args = args[1:]

            (losses, metrics, summaries) = dynamic_args

            # perform updates
            _update_local_context(losses, metrics, summaries)

            # call
            output = f(*args)

            # add outputs context
            return TransformtOutput(
                output=output,
                losses=LOCAL.losses,
                metrics=LOCAL.metrics,
                summaries=LOCAL.summaries,
            )

        # make fn
        transform_fn = transform(
            f_wrapper,
            **kwargs,
        )

        @functools.wraps(f)
        def wrapper(*args):

            # add input context
            dynamic_args = DynamicArgs(
                losses=LOCAL.losses,
                metrics=LOCAL.metrics,
                summaries=LOCAL.summaries,
            )
            # put them first because of static_args
            args = (dynamic_args,) + args

            # call and patch
            (
                output,
                losses,
                metrics,
                summaries,
            ) = transform_fn(*args)

            # perform updates
            _update_local_context(losses, metrics, summaries)

            return output

        return wrapper

    return transform_wrapper


# NOTE: it is unclear if these can be implemented since they dont support `hax_aux`
# jacrev = hooks_aware(jax.jacrev)
# hessian = hooks_aware(jax.hessian)
# mask = hooks_aware(jax.mask)
# jvp = hooks_aware(jax.jvp)
# linearize = hooks_aware(jax.linearize)
# vjp = hooks_aware(jax.vjp)
# linear_transpose = hooks_aware(jax.linear_transpose)
# make_jaxpr = hooks_aware(jax.make_jaxpr)
# defjvp_all = hooks_aware(jax.defjvp_all)
# defjvp = hooks_aware(jax.defjvp)
# defvjp_all = hooks_aware(jax.defvjp_all)
# defvjp = hooks_aware(jax.defvjp)


# NOTE: these can work but require special handling of the axis dimension
# vmap = hooks_aware(jax.vmap)
# pmap = hooks_aware(jax.pmap)
# soft_pmap = hooks_aware(jax.soft_pmap)



import elegy
import jax.numpy as jnp


def test_basic():

    mse = elegy.metrics.MeanSquaredError()

    result = mse.call_with_defaults()(
        y_true=jnp.array([1, 1, 1, 1]), y_pred=jnp.array([0, 1, 1, 1])
    )
    assert result == 0.25

    result = mse.call_with_defaults()(
        y_true=jnp.array([1, 1, 1, 1]), y_pred=jnp.array([1, 0, 0, 0])
    )
    assert result == 0.5



from unittest import TestCase
import numpy as np
import jax.numpy as jnp

import elegy, optax


class DataLoaderTestCase(TestCase):
    def test_basic(self):
        ds = DS0()
        assert len(ds) == 11

        loader = elegy.data.DataLoader(ds, batch_size=4, n_workers=0, shuffle=False)
        batches = list(loader)
        assert len(loader) == len(batches) == 3
        assert len(batches[0]) == 2

        batched_x = [x for x, y in batches]
        batched_y = [y for x, y in batches]

        assert (
            all(batched_y[0] == np.arange(4))
            and all(batched_y[1] == np.arange(4) + 4)
            and all(batched_y[2] == np.arange(3) + 8)
        )
        assert batched_x[0].shape == batched_x[1].shape == (4, 100, 200, 3)
        assert batched_x[2].shape == (3, 100, 200, 3)

    def test_multiprocessed(self):
        ds = DS0()
        loader = elegy.data.DataLoader(ds, batch_size=4, n_workers=4)
        batches = list(loader)
        assert len(loader) == len(batches) == 3
        assert len(batches[0]) == 2

        batched_x = [x for x, y in batches]
        batched_y = [y for x, y in batches]

        assert (
            all(batched_y[0] == np.arange(4))
            and all(batched_y[1] == np.arange(4) + 4)
            and all(batched_y[2] == np.arange(3) + 8)
        )
        assert batched_x[0].shape == batched_x[1].shape == (4, 100, 200, 3)
        assert batched_x[2].shape == (3, 100, 200, 3)

    def test_not_implemented_ds(self):
        class DS0(elegy.data.Dataset):
            ...

        class DS1(elegy.data.Dataset):
            def __len__(self):
                return 5

        for DS in [DS0, DS1]:
            ds = DS()
            loader = elegy.data.DataLoader(ds, 10)
            try:
                list(loader)
            except NotImplementedError:
                pass
            else:
                assert False

    def test_shuffled(self):
        ds = DS0()
        loader = elegy.data.DataLoader(ds, batch_size=4, n_workers=4, shuffle=True)
        batches0 = list(loader)
        batches1 = list(loader)

        batched_y0 = [y for x, y in batches0]
        batched_y1 = [y for x, y in batches1]

        assert not (
            all(batched_y0[0] == np.arange(4))
            and all(batched_y0[1] == np.arange(4) + 4)
            and all(batched_y0[2] == np.arange(3) + 8)
        )
        assert not (
            all(batched_y0[0] == batched_y1[0])
            and all(batched_y0[1] == batched_y1[1])
            and all(batched_y0[2] == batched_y1[2])
        )

    def test_model_fit(self):
        ds = DS0()
        loader_train = elegy.data.DataLoader(ds, batch_size=4, n_workers=4)
        loader_valid = elegy.data.DataLoader(ds, batch_size=4, n_workers=4)

        class Module(elegy.Module):
            def call(self, x):
                x = jnp.mean(x, axis=[1, 2])
                x = elegy.nn.Linear(20)(x)
                return x

        model = elegy.Model(
            Module(),
            loss=elegy.losses.SparseCategoricalCrossentropy(),
            optimizer=optax.sgd(0.1),
        )
        model.fit(loader_train, validation_data=loader_valid, epochs=3)

    def test_multi_item_ds(self):
        # one item
        class DS0:
            def __len__(self):
                return 11

            def __getitem__(self, i):
                return np.zeros([100, 200, 3])

        loader0 = elegy.data.DataLoader(DS0(), batch_size=4)
        batches0 = list(loader0)
        assert len(batches0) == 3
        assert (
            batches0[0].shape == (4, 100, 200, 3)
            and batches0[1].shape == (4, 100, 200, 3)
            and batches0[2].shape == (3, 100, 200, 3)
        )

        # three items
        class DS1:
            def __len__(self):
                return 11

            def __getitem__(self, i):
                return (
                    np.zeros([100, 200, 3]),
                    np.zeros([100, 200, 3]),
                    np.zeros([100, 200, 3]),
                )

        loader1 = elegy.data.DataLoader(DS1(), batch_size=4)
        batches1 = list(loader1)
        assert len(batches1) == 3
        assert isinstance(batches1[0], tuple)
        assert len(batches1[0]) == 3
        assert (
            batches1[0][0].shape == (4, 100, 200, 3)
            and batches1[1][1].shape == (4, 100, 200, 3)
            and batches1[2][2].shape == (3, 100, 200, 3)
        )

    def test_worker_type(self):
        ds = DS0()
        for worker_type in ["thread", "process"]:
            loader = elegy.data.DataLoader(
                ds, batch_size=4, n_workers=4, worker_type=worker_type
            )
            batches = list(loader)

    def test_prefetch(self):
        ds = DS0()

        loader = elegy.data.DataLoader(
            ds, batch_size=2, n_workers=4, shuffle=False, prefetch=3
        )
        batches = list(loader)
        assert len(loader) == len(batches)

    def test_custom_batch_fn(self):
        ds = DS_custom_batch_fn()
        loader = elegy.data.DataLoader(ds, batch_size=3)
        batches = list(loader)
        assert len(loader) == len(batches)
        assert batches[0]["a"].shape == (3, 10)
        assert batches[0]["b"].shape == (3,)
        assert batches[0]["c"] == "banana"
        assert np.all(batches[0]["b"] == np.array([0, 1, 2]))

    def test_loader_from_array(self):
        pseudo_ds = np.arange(65)
        loader = elegy.data.DataLoader(pseudo_ds, batch_size=10)
        batches = list(loader)
        assert len(batches) == 7
        assert np.all(batches[1] == np.arange(10, 20))


class DS0(elegy.data.Dataset):
    def __len__(self):
        return 11

    def __getitem__(self, i):
        return np.zeros([100, 200, 3]), np.arange(20)[i]


class DS_custom_batch_fn(elegy.data.Dataset):
    def __len__(self):
        return 11

    def __getitem__(self, i):
        return dict(a=np.random.random(size=10), b=i)

    def batch_fn(self, list_of_samples):
        x = super().batch_fn(list_of_samples)
        x.update(c="banana")
        return x



from dataclasses import dataclass
import sys
import typing as tp
from copy import copy
from enum import Enum
from functools import total_ordering

import jax
import jax.numpy as jnp
import jax.tree_util
import numpy as np


if sys.version_info >= (3, 8):
    from typing import Protocol, runtime_checkable
else:
    from typing_extensions import Protocol, runtime_checkable

EPSILON = 1e-7
F = tp.TypeVar("F", bound=tp.Callable)


class TrivialPytree:
    def tree_flatten(self):
        return tuple(vars(self).values()), None

    @classmethod
    def tree_unflatten(cls, _aux_data, children):
        return cls(*children)


class Empty:
    pass


EMPTY = Empty()


@jax.tree_util.register_pytree_node_class
class RNGSeq(TrivialPytree):
    key: jnp.ndarray

    def __init__(self, key: tp.Union[int, jnp.ndarray]):
        self.key = (
            jax.random.PRNGKey(key)
            if isinstance(key, int) or (hasattr(key, "shape") and key.shape == ())
            else key
        )

    def next(self) -> jnp.ndarray:
        self.key = jax.random.split(self.key, 1)[0]
        return self.key

    def __repr__(self) -> str:
        return f"RNGSeq(key={self.key})"


T = tp.TypeVar("T")
Container = tp.Union[
    T,
    tp.Tuple["Container", ...],
    tp.Dict[str, "Container"],
]
ArrayHolder = tp.Union[Container[np.ndarray], np.ndarray]

IndexLike = tp.Union[str, int, tp.Iterable[tp.Union[str, int]]]

Shape = tp.Sequence[int]
ShapeLike = tp.Union[int, Shape]
FloatLike = tp.Union[float, np.ndarray]
DType = tp.Any
ParamName = str

Params = tp.Mapping[str, tp.Mapping[ParamName, np.ndarray]]
State = tp.Mapping[str, tp.Mapping[str, np.ndarray]]
PadFn = tp.Callable[[int], tp.Tuple[int, int]]
PadFnOrFns = tp.Union[PadFn, tp.Sequence[PadFn]]
PRNGKey = np.ndarray
Parameters = tp.Dict[str, tp.Any]
ParameterCollection = tp.Dict[str, Parameters]
Logs = tp.Dict[str, tp.Union[np.ndarray, float]]
Index = tp.Union[int, str]
Path = tp.Tuple[Index, ...]
Grads = tp.Any
RNG = tp.Union[RNGSeq, np.ndarray]
Scalar = tp.Union[np.ndarray, float, int]
SummaryModule = tp.Any
SummaryValue = tp.Any

NetParams = tp.Any
NetStates = tp.Any
ModuleParams = tp.Any
ModuleStates = tp.Any
MetricsStates = tp.Any
OptimizerStates = tp.Any
OptimizerStates = tp.Any
Grads = tp.Any
Pytree = tp.Any


@jax.tree_util.register_pytree_node_class
class Summary(tp.NamedTuple):
    path: Path
    module: tp.Optional[SummaryModule]
    value: SummaryValue

    def tree_flatten(self):
        return ((self.value,), (self.path, self.module))

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        (value,) = children
        path, module = aux_data

        return cls(path, module, value)


Summaries = tp.List[Summary]


@jax.tree_util.register_pytree_node_class
class SummaryTableEntry(tp.NamedTuple):
    path: str
    module_type_name: str
    output_value: Pytree
    trainable_params_count: int
    trainable_params_size: int
    non_trainable_params_count: int
    non_trainable_params_size: int

    @classmethod
    def totals_entry(
        cls,
        trainable_params_count: int,
        trainable_params_size: int,
        non_trainable_params_count: int,
        non_trainable_params_size: int,
    ):
        return cls(
            path="",
            module_type_name="",
            output_value=None,
            trainable_params_count=trainable_params_count,
            trainable_params_size=trainable_params_size,
            non_trainable_params_count=non_trainable_params_count,
            non_trainable_params_size=non_trainable_params_size,
        )

    def tree_flatten(self):
        return (
            (self.output_value,),
            (
                self.path,
                self.module_type_name,
                self.trainable_params_count,
                self.trainable_params_size,
                self.non_trainable_params_count,
                self.non_trainable_params_size,
            ),
        )

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        (
            path,
            module_type_name,
            trainable_params_count,
            trainable_params_size,
            non_trainable_params_count,
            non_trainable_params_size,
        ) = aux_data
        (output_value,) = children

        return cls(
            path=path,
            module_type_name=module_type_name,
            output_value=output_value,
            trainable_params_count=trainable_params_count,
            trainable_params_size=trainable_params_size,
            non_trainable_params_count=non_trainable_params_count,
            non_trainable_params_size=non_trainable_params_size,
        )


class OutputStates(tp.NamedTuple):
    preds: tp.Any
    params: tp.Any
    states: tp.Any


@jax.tree_util.register_pytree_node_class
class States(tp.Mapping):
    def __init__(self, _data=None, **kwargs):
        self.__dict__.update(
            dict(_data, **kwargs) if _data is not None else dict(**kwargs)
        )

    def __len__(self) -> int:
        return len(self.__dict__)

    def __getitem__(self, key):
        return self.__dict__[key]

    def __iter__(self):
        return iter(self.__dict__)

    def __getattr__(self, key):
        try:
            return object.__getattr__(self, key)
        except AttributeError:
            raise AttributeError(
                f"'{self.__class__.__name__}' has not attribute '{key}'"
            )

    def __setattr__(self, key, value):
        raise AttributeError("can't set attribute")

    def update(self, **kwargs) -> "States":
        """Returns a new States object, updating all attributes from kwargs."""
        data = self.__dict__.copy()
        data.update(kwargs)
        return States(data)

    def maybe_update(self, **kwargs) -> "States":
        """Returns a new States object, updating attributes that are not yet present."""
        kwargs = {
            key: value
            for key, value in kwargs.items()
            if key not in self.__dict__ or self.__dict__[key] is None
        }

        return self.update(**kwargs)

    def update_known(*self, **kwargs) -> "States":
        """Returns a new States object, updating attributes that are already present.
        e.g: states.update_known(**locals())"""
        # NOTE: first argument is *self to allow the **locals() syntax inside bound methods
        # which have their own self inside locals()
        # otherwise will get a "got multiple values for argument 'self'" error"
        assert len(self) == 1, "States.update_known() called with positional arguments"
        self = self[0]

        kwargs = {key: value for key, value in kwargs.items() if key in self.__dict__}
        return self.update(**kwargs)

    def copy(self) -> "States":
        return jax.tree_map(lambda x: x, self)

    def tree_flatten(self):
        return (tuple(self.__dict__.values()), tuple(self.__dict__.keys()))

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        return cls(zip(aux_data, children))


@dataclass
class Parameter:
    collection: str
    value: tp.Any


@dataclass
class Info:
    shape: tp.Tuple[int, ...]
    dtype: tp.Any


@dataclass
class ParameterSpec:
    collection: str
    info: tp.Any


class Initializer(Protocol):
    def __call__(self, shape: Shape, dtype: DType) -> np.ndarray:
        ...


class JitCallable(Protocol):
    def __call__(
        self, *args
    ) -> tp.Tuple[tp.Any, tp.Optional[Parameters], ParameterCollection]:
        ...


class InitJit(Protocol):
    def __call__(
        self,
        *,
        rng: tp.Optional[RNGSeq] = None,
        set_defaults: bool = False,
    ) -> JitCallable:
        ...


class ApplyJit(Protocol):
    def __call__(
        self,
        params: tp.Optional[Parameters],
        collections: ParameterCollection,
        *,
        training: bool = True,
        rng: tp.Optional[RNGSeq] = None,
        set_defaults: bool = False,
    ) -> JitCallable:
        ...


class MissingModule(Exception):
    pass


class MissingOptimizer(Exception):
    pass


class MissingMethod(Exception):
    pass


class DependencyUnavailable(Exception):
    pass


class ShapeMismatch(Exception):
    pass


class MissingParameter(Exception):
    pass


class NoContext(Exception):
    pass


class ModuleOrderError(Exception):
    pass


class SubmoduleNotRegistered(Exception):
    pass


class ModelNotInitialized(Exception):
    pass



# Implementation based on tf.keras.callbacks.py
# https://github.com/tensorflow/tensorflow/blob/v2.2.0/tensorflow/python/keras/callbacks.py
try:
    import requests
except ImportError:
    requests = None
import json
import logging
import typing as tp

import numpy as np

from .callback import Callback


class RemoteMonitor(Callback):
    """Callback used to stream events to a server.

    Requires the `requests` library.
    Events are sent to `root + '/publish/epoch/end/'` by default. Calls are
    HTTP POST, with a `data` argument which is a
    JSON-encoded dictionary of event data.
    If send_as_json is set to True, the content type of the request will be
    application/json. Otherwise the serialized JSON will be sent within a form.
    """

    def __init__(
        self,
        root: str = "http://localhost:9000",
        path: str = "/publish/epoch/end/",
        field: str = "data",
        headers: tp.Optional[tp.Dict[str, str]] = None,
        send_as_json: bool = False,
    ):
        """
        Arguments:
            root: String; root url of the target server.
            path: String; path relative to `root` to which the events will be sent.
            field: String; JSON field under which the data will be stored.
                The field is used only if the payload is sent within a form
                (i.e. send_as_json is set to False).
            headers: Dictionary; optional custom HTTP headers.
            send_as_json: Boolean; whether the request should be
                sent as application/json.
        """
        super(RemoteMonitor, self).__init__()

        self.root = root
        self.path = path
        self.field = field
        self.headers = headers
        self.send_as_json = send_as_json

    def on_epoch_end(self, epoch, logs=None):
        if requests is None:
            raise ImportError("RemoteMonitor requires the `requests` library.")
        logs = logs or {}
        send = {}
        send["epoch"] = epoch
        for k, v in logs.items():
            # np.ndarray and np.generic are not scalar types
            # therefore we must unwrap their scalar values and
            # pass to the json-serializable dict 'send'
            if isinstance(v, (np.ndarray, np.generic)):
                send[k] = v.item()
            else:
                send[k] = v
        try:
            if self.send_as_json:
                requests.post(self.root + self.path, json=send, headers=self.headers)
            else:
                requests.post(
                    self.root + self.path,
                    {self.field: json.dumps(send)},
                    headers=self.headers,
                )
        except requests.exceptions.RequestException:
            logging.warning(
                "Warning: could not reach RemoteMonitor "
                "root server at " + str(self.root)
            )



import math
from unittest import TestCase

import numpy as np
import jax.numpy as jnp

import pytest

from elegy.data import utils


class TrainValidationSplitTest(TestCase):
    def test_basic(self):
        x_all = np.random.uniform(size=(100, 32, 32, 3))
        y_all = np.random.uniform(size=(100, 1))
        sample_weight_all = None
        split = 0.2

        (x, y, sample_weight), validation_data = utils.train_validation_split(
            (x_all, y_all, sample_weight_all), validation_split=0.2, shuffle=False
        )

        assert x.shape[0] == int(x_all.shape[0] * (1 - split))
        assert y.shape[0] == int(y_all.shape[0] * (1 - split))
        assert sample_weight is None

        (x, y, sample_weight) = validation_data
        assert x.shape[0] == int(x_all.shape[0] * split)
        assert y.shape[0] == int(y_all.shape[0] * split)
        assert sample_weight is None



import typing as tp
from enum import Enum

import jax.numpy as jnp
import numpy as np

from elegy import initializers, module, types, utils
from elegy.metrics.metric import Metric


class Reduction(Enum):
    SUM = "sum"
    SUM_OVER_BATCH_SIZE = "sum_over_batch_size"
    WEIGHTED_MEAN = "weighted_mean"


def reduce(
    total: jnp.ndarray,
    count: tp.Optional[jnp.ndarray],
    values: jnp.ndarray,
    reduction: Reduction,
    sample_weight: tp.Optional[np.ndarray],
    dtype: jnp.dtype,
) -> tp.Tuple[jnp.ndarray, jnp.ndarray, tp.Optional[jnp.ndarray]]:

    if sample_weight is not None:
        sample_weight = sample_weight.astype(dtype)

        # Update dimensions of weights to match with values if possible.
        # values, _, sample_weight = tf_losses_utils.squeeze_or_expand_dimensions(
        #     values, sample_weight=sample_weight
        # )

        try:
            # Broadcast weights if possible.
            sample_weight = jnp.broadcast_to(sample_weight, values.shape)
        except ValueError:
            # Reduce values to same ndim as weight array
            ndim = values.ndim
            weight_ndim = sample_weight.ndim
            if reduction == Reduction.SUM:
                values = jnp.sum(values, axis=list(range(weight_ndim, ndim)))
            else:
                values = jnp.mean(values, axis=list(range(weight_ndim, ndim)))

        values = values * sample_weight

    value_sum = jnp.sum(values)

    total += value_sum

    # Exit early if the reduction doesn't have a denominator.
    if reduction == Reduction.SUM:
        num_values = None

    # Update `count` for reductions that require a denominator.
    elif reduction == Reduction.SUM_OVER_BATCH_SIZE:
        num_values = jnp.prod(values.shape).astype(dtype)

    else:
        if sample_weight is None:
            num_values = jnp.prod(jnp.array(values.shape)).astype(dtype)
        else:
            num_values = jnp.sum(sample_weight)

    if count is not None and num_values is not None:
        count += num_values

    if reduction == Reduction.SUM:
        value = total
    else:
        value = total / count

    return value, total, count


class Reduce(Metric):
    """Encapsulates metrics that perform a reduce operation on the values."""

    def __init__(self, reduction: Reduction, **kwargs):
        super().__init__(**kwargs)

        self._reduction = reduction

        if self._reduction not in (
            Reduction.SUM,
            Reduction.SUM_OVER_BATCH_SIZE,
            Reduction.WEIGHTED_MEAN,
        ):
            raise NotImplementedError(
                "reduction {reduction} not implemented".format(
                    reduction=self._reduction
                )
            )

    def call(
        self, values: jnp.ndarray, sample_weight: tp.Optional[jnp.ndarray] = None
    ) -> jnp.ndarray:
        """
        Accumulates statistics for computing the reduction metric. For example, if `values` is [1, 3, 5, 7]
        and reduction=SUM_OVER_BATCH_SIZE, then the value of `result()` is 4. If the `sample_weight`
        is specified as [1, 1, 0, 0] then value of `result()` would be 2.

        Arguments:
            values: Per-example value.
            sample_weight: Optional weighting of each example. Defaults to 1.

        Returns:
            Array with the cumulative reduce.
        """
        total = self.add_parameter(
            "total",
            lambda: jnp.array(0, dtype=jnp.int32),
            trainable=False,
        )

        if self._reduction in (
            Reduction.SUM_OVER_BATCH_SIZE,
            Reduction.WEIGHTED_MEAN,
        ):
            count = self.add_parameter(
                "count",
                lambda: jnp.array(0, dtype=jnp.int32),
                trainable=False,
            )
        else:
            count = None

        value, total, count = reduce(
            total=total,
            count=count,
            values=values,
            reduction=self._reduction,
            sample_weight=sample_weight,
            dtype=self.dtype,
        )

        self.update_parameter("total", total)

        if count is not None:
            self.update_parameter("count", count)

        return value



# Lint as: python3
# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Convolutional Haiku modules."""

import typing as tp

import haiku as hk
import jax.numpy as jnp
import numpy as np
from haiku._src import utils as hk_utils
from jax import lax

from elegy import initializers, module, types


def to_dimension_numbers(
    num_spatial_dims: int,
    channels_last: bool,
    transpose: bool,
) -> lax.ConvDimensionNumbers:
    """Create a `lax.ConvDimensionNumbers` for the given inputs."""
    num_dims = num_spatial_dims + 2

    if channels_last:
        spatial_dims = tuple(range(1, num_dims - 1))
        image_dn = (0, num_dims - 1) + spatial_dims
    else:
        spatial_dims = tuple(range(2, num_dims))
        image_dn = (0, 1) + spatial_dims

    if transpose:
        kernel_dn = (num_dims - 2, num_dims - 1) + tuple(range(num_dims - 2))
    else:
        kernel_dn = (num_dims - 1, num_dims - 2) + tuple(range(num_dims - 2))

    return lax.ConvDimensionNumbers(
        lhs_spec=image_dn, rhs_spec=kernel_dn, out_spec=image_dn
    )


class ConvND(module.Module):
    """General N-dimensional convolutional."""

    def __init__(
        self,
        num_spatial_dims: int,
        output_channels: int,
        kernel_shape: tp.Union[int, tp.Sequence[int]],
        stride: tp.Union[int, tp.Sequence[int]] = 1,
        rate: tp.Union[int, tp.Sequence[int]] = 1,
        padding: tp.Union[
            str, tp.Sequence[tp.Tuple[int, int]], types.PadFnOrFns
        ] = "SAME",
        with_bias: bool = True,
        w_init: tp.Optional[types.Initializer] = None,
        b_init: tp.Optional[types.Initializer] = None,
        data_format: str = "channels_last",
        mask: tp.Optional[np.ndarray] = None,
        groups: int = 1,
        **kwargs,
    ):
        """
        Initializes the module.

        Args:
            num_spatial_dims: The number of spatial dimensions of the input.
            output_channels: Number of output channels.
            kernel_shape: The shape of the kernel. Either an integer or a sequence of
                length ``num_spatial_dims``.
            stride: tp.Optional stride for the kernel. Either an integer or a sequence of
                length ``num_spatial_dims``. Defaults to 1.
            rate: tp.Optional kernel dilation rate. Either an integer or a sequence of
                length ``num_spatial_dims``. 1 corresponds to standard ND convolution,
                ``rate > 1`` corresponds to dilated convolution. Defaults to 1.
            padding: tp.Optional padding algorithm. Either ``VALID`` or ``SAME`` or a
                sequence of n ``(low, high)`` integer pairs that give the padding to
                apply before and after each spatial dimension. or a callable or sequence
                of callables of size ``num_spatial_dims``. Any callables must take a
                single integer argument equal to the effective kernel size and return a
                sequence of two integers representing the padding before and after. See
                ``haiku.pad.*`` for more details and example functions. Defaults to
                ``SAME``. See:
                https://www.tensorflow.org/xla/operation_semantics#conv_convolution.
            with_bias: Whether to add a bias. By default, true.
            w_init: tp.Optional weight initialization. By default, truncated normal.
            b_init: tp.Optional bias initialization. By default, zeros.
            data_format: The data format of the input.  Can be either
                ``channels_first``, ``channels_last``, ``N...C`` or ``NC...``. By
                default, ``channels_last``.
            mask: tp.Optional mask of the weights.
            groups: A positive integer specifying the number of groups in which the
                input is split along the channel axis. Each group is convolved separately
                with filters / groups filters. The output is the concatenation of all the
                groups results along the channel axis. Input channels and filters must both
                be divisible by groups.
            kwargs: Additional keyword arguments passed to Module.
        """
        super().__init__(**kwargs)
        if num_spatial_dims <= 0:
            raise ValueError(
                "We only support convolution operations for `num_spatial_dims` "
                f"greater than 0, received num_spatial_dims={num_spatial_dims}."
            )

        self.num_spatial_dims = num_spatial_dims
        self.output_channels = output_channels
        self.kernel_shape = hk_utils.replicate(
            kernel_shape, num_spatial_dims, "kernel_shape"
        )
        self.with_bias = with_bias
        self.stride = hk_utils.replicate(stride, num_spatial_dims, "strides")
        self.w_init = w_init
        self.b_init = b_init or jnp.zeros
        self.mask = mask
        self.lhs_dilation = hk_utils.replicate(1, num_spatial_dims, "lhs_dilation")
        self.kernel_dilation = hk_utils.replicate(
            rate, num_spatial_dims, "kernel_dilation"
        )
        self.data_format = data_format
        self.channel_index = hk_utils.get_channel_index(data_format)
        self.dimension_numbers = to_dimension_numbers(
            num_spatial_dims, channels_last=(self.channel_index == -1), transpose=False
        )
        self.groups = groups

        if isinstance(padding, str):
            self.padding = padding.upper()
        else:
            self.padding = hk.pad.create(
                padding=padding,
                kernel=self.kernel_shape,
                rate=self.kernel_dilation,
                n=self.num_spatial_dims,
            )

    def call(self, inputs: np.ndarray) -> np.ndarray:
        """
        Connects ``ConvND`` layer.

        Args:
            inputs: A rank-N+2 array with shape ``[N, spatial_dims, C]``.

        Returns:
            A rank-N+2 array with shape ``[N, spatial_dims, output_channels]``.
        """
        required_rank = self.num_spatial_dims + 2
        if inputs.ndim != required_rank:
            raise ValueError(
                f"Input to ConvND needs to have rank {required_rank}, "
                f"but input has shape {inputs.shape}."
            )

        w_shape = self.kernel_shape + (
            inputs.shape[self.channel_index] // self.groups,
            self.output_channels,
        )

        if self.mask is not None and self.mask.shape != w_shape:
            raise ValueError(
                "Mask needs to have the same shape as weights. "
                f"Shapes are: {self.mask.shape}, {w_shape}"
            )

        w_init = self.w_init
        if w_init is None:
            fan_in_shape = np.prod(w_shape[:-1])
            stddev = 1.0 / np.sqrt(fan_in_shape)
            w_init = initializers.TruncatedNormal(stddev=stddev)
        w = self.add_parameter("w", lambda: w_init(w_shape, jnp.float32))

        if self.mask is not None:
            w *= self.mask

        inputs = jnp.asarray(inputs, dtype=self.dtype)
        w = jnp.asarray(w, dtype=self.dtype)
        out = lax.conv_general_dilated(
            inputs,
            w,
            window_strides=self.stride,
            padding=self.padding,
            lhs_dilation=self.lhs_dilation,
            rhs_dilation=self.kernel_dilation,
            dimension_numbers=self.dimension_numbers,
            feature_group_count=self.groups,
        )

        if self.with_bias:
            if self.channel_index == -1:
                bias_shape = (self.output_channels,)
            else:
                bias_shape = (self.output_channels,) + (1,) * self.num_spatial_dims
            b = self.add_parameter("b", lambda: self.b_init(bias_shape, jnp.float32))
            b = jnp.broadcast_to(b, out.shape)
            b = jnp.asarray(b, self.dtype)
            out = out + b

        return out


class Conv1D(ConvND):
    """One dimensional convolution."""

    def __init__(
        self,
        output_channels: int,
        kernel_shape: tp.Union[int, tp.Sequence[int]],
        stride: tp.Union[int, tp.Sequence[int]] = 1,
        rate: tp.Union[int, tp.Sequence[int]] = 1,
        padding: tp.Union[
            str, tp.Sequence[tp.Tuple[int, int]], types.PadFnOrFns
        ] = "SAME",
        with_bias: bool = True,
        w_init: tp.Optional[types.Initializer] = None,
        b_init: tp.Optional[types.Initializer] = None,
        data_format: str = "NWC",
        mask: tp.Optional[np.ndarray] = None,
        groups: int = 1,
        **kwargs,
    ):
        """
        Initializes the module.

        Args:
            output_channels: Number of output channels.
            kernel_shape: The shape of the kernel. Either an integer or a sequence of
                length 1.
            stride: tp.Optional stride for the kernel. Either an integer or a sequence of
                length 1. Defaults to 1.
            rate: tp.Optional kernel dilation rate. Either an integer or a sequence of
                length 1. 1 corresponds to standard ND convolution,
                ``rate > 1`` corresponds to dilated convolution. Defaults to 1.
            padding: tp.Optional padding algorithm. Either ``VALID`` or ``SAME`` or
                a callable or sequence of callables of length 1. Any callables must
                take a single integer argument equal to the effective kernel size and
                return a list of two integers representing the padding before and after.
                See haiku.pad.* for more details and example functions.
                Defaults to ``SAME``. See:
                https://www.tensorflow.org/xla/operation_semantics#conv_convolution.
            with_bias: Whether to add a bias. By default, true.
            w_init: tp.Optional weight initialization. By default, truncated normal.
            b_init: tp.Optional bias initialization. By default, zeros.
            data_format: The data format of the input. Either ``NWC`` or ``NCW``. By
                default, ``NWC``.
            mask: tp.Optional mask of the weights.
            groups: A positive integer specifying the number of groups in which the
                input is split along the channel axis. Each group is convolved separately
                with filters / groups filters. The output is the concatenation of all the
                groups results along the channel axis. Input channels and filters must both
                be divisible by groups.
            kwargs: Additional keyword arguments passed to Module.
        """
        super().__init__(
            num_spatial_dims=1,
            output_channels=output_channels,
            kernel_shape=kernel_shape,
            stride=stride,
            rate=rate,
            padding=padding,
            with_bias=with_bias,
            w_init=w_init,
            b_init=b_init,
            data_format=data_format,
            mask=mask,
            groups=groups,
            **kwargs,
        )


class Conv2D(ConvND):
    """Two dimensional convolution."""

    def __init__(
        self,
        output_channels: int,
        kernel_shape: tp.Union[int, tp.Sequence[int]],
        stride: tp.Union[int, tp.Sequence[int]] = 1,
        rate: tp.Union[int, tp.Sequence[int]] = 1,
        padding: tp.Union[
            str, tp.Sequence[tp.Tuple[int, int]], types.PadFnOrFns
        ] = "SAME",
        with_bias: bool = True,
        w_init: tp.Optional[types.Initializer] = None,
        b_init: tp.Optional[types.Initializer] = None,
        data_format: str = "NHWC",
        mask: tp.Optional[np.ndarray] = None,
        groups: int = 1,
        **kwargs,
    ):
        """
        Initializes the module.

        Args:
            output_channels: Number of output channels.
            kernel_shape: The shape of the kernel. Either an integer or a sequence of
                length 2.
            stride: tp.Optional stride for the kernel. Either an integer or a sequence of
                length 2. Defaults to 1.
            rate: tp.Optional kernel dilation rate. Either an integer or a sequence of
                length 2. 1 corresponds to standard ND convolution,
                ``rate > 1`` corresponds to dilated convolution. Defaults to 1.
            padding: tp.Optional padding algorithm. Either ``VALID`` or ``SAME`` or
                a callable or sequence of callables of length 2. Any callables must
                take a single integer argument equal to the effective kernel size and
                return a list of two integers representing the padding before and after.
                See haiku.pad.* for more details and example functions.
                Defaults to ``SAME``. See:
                https://www.tensorflow.org/xla/operation_semantics#conv_convolution.
            with_bias: Whether to add a bias. By default, true.
            w_init: tp.Optional weight initialization. By default, truncated normal.
            b_init: tp.Optional bias initialization. By default, zeros.
            data_format: The data format of the input. Either ``NHWC`` or ``NCHW``. By
                default, ``NHWC``.
            mask: tp.Optional mask of the weights.
            groups: A positive integer specifying the number of groups in which the
                input is split along the channel axis. Each group is convolved separately
                with filters / groups filters. The output is the concatenation of all the
                groups results along the channel axis. Input channels and filters must both
                be divisible by groups.
            kwargs: Additional keyword arguments passed to Module.
        """
        super().__init__(
            num_spatial_dims=2,
            output_channels=output_channels,
            kernel_shape=kernel_shape,
            stride=stride,
            rate=rate,
            padding=padding,
            with_bias=with_bias,
            w_init=w_init,
            b_init=b_init,
            data_format=data_format,
            mask=mask,
            groups=groups,
            **kwargs,
        )


class Conv3D(ConvND):
    """Three dimensional convolution."""

    def __init__(
        self,
        output_channels: int,
        kernel_shape: tp.Union[int, tp.Sequence[int]],
        stride: tp.Union[int, tp.Sequence[int]] = 1,
        rate: tp.Union[int, tp.Sequence[int]] = 1,
        padding: tp.Union[
            str, tp.Sequence[tp.Tuple[int, int]], types.PadFnOrFns
        ] = "SAME",
        with_bias: bool = True,
        w_init: tp.Optional[types.Initializer] = None,
        b_init: tp.Optional[types.Initializer] = None,
        data_format: str = "NDHWC",
        mask: tp.Optional[np.ndarray] = None,
        groups: int = 1,
        **kwargs,
    ):
        """
        Initializes the module.

        Args:
            output_channels: Number of output channels.
            kernel_shape: The shape of the kernel. Either an integer or a sequence of
                length 3.
            stride: tp.Optional stride for the kernel. Either an integer or a sequence of
                length 3. Defaults to 1.
            rate: tp.Optional kernel dilation rate. Either an integer or a sequence of
                length 3. 1 corresponds to standard ND convolution,
                `rate > 1` corresponds to dilated convolution. Defaults to 1.
            padding: tp.Optional padding algorithm. Either ``VALID`` or ``SAME`` or
                a callable or sequence of callables of length 3. Any callables must
                take a single integer argument equal to the effective kernel size and
                return a list of two integers representing the padding before and after.
                See haiku.pad.* for more details and example functions.
                Defaults to ``SAME``. See:
                https://www.tensorflow.org/xla/operation_semantics#conv_convolution.
            with_bias: Whether to add a bias. By default, true.
            w_init: tp.Optional weight initialization. By default, truncated normal.
            b_init: tp.Optional bias initialization. By default, zeros.
            data_format: The data format of the input. Either ``NDHWC`` or ``NCDHW``.
                By default, ``NDHWC``.
            mask: tp.Optional mask of the weights.
            groups: A positive integer specifying the number of groups in which the
                input is split along the channel axis. Each group is convolved separately
                with filters / groups filters. The output is the concatenation of all the
                groups results along the channel axis. Input channels and filters must both
                be divisible by groups.
            kwargs: Additional keyword arguments passed to Module.
        """
        super().__init__(
            num_spatial_dims=3,
            output_channels=output_channels,
            kernel_shape=kernel_shape,
            stride=stride,
            rate=rate,
            padding=padding,
            with_bias=with_bias,
            w_init=w_init,
            b_init=b_init,
            data_format=data_format,
            mask=mask,
            groups=groups,
            **kwargs,
        )


class ConvNDTranspose(module.Module):
    """General n-dimensional transposed convolution (aka. deconvolution)."""

    def __init__(
        self,
        num_spatial_dims: int,
        output_channels: int,
        kernel_shape: tp.Union[int, tp.Sequence[int]],
        stride: tp.Union[int, tp.Sequence[int]] = 1,
        padding: tp.Union[str, tp.Sequence[tp.Tuple[int, int]]] = "SAME",
        with_bias: bool = True,
        w_init: tp.Optional[types.Initializer] = None,
        b_init: tp.Optional[types.Initializer] = None,
        data_format: str = "channels_last",
        mask: tp.Optional[np.ndarray] = None,
        **kwargs,
    ):
        """
        Initializes the module.

        Args:
            num_spatial_dims: The number of spatial dimensions of the input.
            output_channels: Number of output channels.
            kernel_shape: The shape of the kernel. Either an integer or a sequence of
                length ``num_spatial_dims``.
            stride: tp.Optional stride for the kernel. Either an integer or a sequence of
                length ``num_spatial_dims``. Defaults to 1.
            padding: tp.Optional padding algorithm. Either "VALID" or "SAME".
                Defaults to "SAME". See:
                https://www.tensorflow.org/xla/operation_semantics#conv_convolution.
            with_bias: Whether to add a bias. By default, true.
            w_init: tp.Optional weight initialization. By default, truncated normal.
            b_init: tp.Optional bias initialization. By default, zeros.
            data_format: The data format of the input. Can be either
                ``channels_first``, ``channels_last``, ``N...C`` or ``NC...``. By
                default, ``channels_last``.
            mask: tp.Optional mask of the weights.
            kwargs: Additional keyword arguments passed to Module.
        """
        super().__init__(**kwargs)

        if num_spatial_dims <= 0:
            raise ValueError(
                "We only support convolution operations for `num_spatial_dims` "
                f"greater than 0, received num_spatial_dims={num_spatial_dims}."
            )

        self.num_spatial_dims = num_spatial_dims
        self.output_channels = output_channels
        self.kernel_shape = hk_utils.replicate(
            kernel_shape, num_spatial_dims, "kernel_shape"
        )
        self.with_bias = with_bias
        self.stride = hk_utils.replicate(stride, num_spatial_dims, "strides")
        self.w_init = w_init
        self.b_init = b_init or jnp.zeros
        self.mask = mask
        # TODO(tomhennigan) Make use of hk.pad.create here?
        self.padding = padding
        self.data_format = data_format
        self.channel_index = hk_utils.get_channel_index(data_format)
        self.dimension_numbers = to_dimension_numbers(
            num_spatial_dims, channels_last=(self.channel_index == -1), transpose=True
        )

    def call(self, inputs: np.ndarray) -> np.ndarray:
        """
        Computes the transposed convolution of the input.

        Args:
            inputs: A rank-N+2 array with shape ``[N, spatial_dims, C]``.

        Returns:
            A rank-N+2 array with shape ``[N, spatial_dims, output_channels]``.
        """
        required_rank = self.num_spatial_dims + 2
        if inputs.ndim != required_rank:
            raise ValueError(
                f"Input to ConvND needs to have rank {required_rank}, "
                f"but input has shape {inputs.shape}."
            )

        input_channels = inputs.shape[self.channel_index]
        w_shape = self.kernel_shape + (self.output_channels, input_channels)

        if self.mask is not None and self.mask.shape != w_shape:
            raise ValueError(
                "Mask needs to have the same shape as weights. "
                f"Shapes are: {self.mask.shape}, {w_shape}"
            )

        w_init = self.w_init
        if w_init is None:
            fan_in_shape = self.kernel_shape + (input_channels,)
            stddev = 1.0 / np.sqrt(np.prod(fan_in_shape))
            w_init = initializers.TruncatedNormal(stddev=stddev)
        w = self.add_parameter("w", lambda: w_init(w_shape, inputs.dtype))

        if self.mask is not None:
            w = w * self.mask

        out = lax.conv_transpose(
            inputs,
            w,
            strides=self.stride,
            padding=self.padding,
            dimension_numbers=self.dimension_numbers,
        )

        if self.with_bias:
            if self.channel_index == -1:
                bias_shape = (self.output_channels,)
            else:
                bias_shape = (self.output_channels,) + (1,) * self.num_spatial_dims
            b = self.add_parameter("b", lambda: self.b_init(bias_shape, inputs.dtype))
            b = jnp.broadcast_to(b, out.shape)
            out = out + b

        return out


class Conv1DTranspose(ConvNDTranspose):
    """One dimensional transposed convolution (aka. deconvolution)."""

    def __init__(
        self,
        output_channels: int,
        kernel_shape: tp.Union[int, tp.Sequence[int]],
        stride: tp.Union[int, tp.Sequence[int]] = 1,
        padding: tp.Union[str, tp.Sequence[tp.Tuple[int, int]]] = "SAME",
        with_bias: bool = True,
        w_init: tp.Optional[types.Initializer] = None,
        b_init: tp.Optional[types.Initializer] = None,
        data_format: str = "NWC",
        mask: tp.Optional[np.ndarray] = None,
        **kwargs,
    ):
        """
        Initializes the module.

        Args:
            output_channels: Number of output channels.
            kernel_shape: The shape of the kernel. Either an integer or a sequence of
                length 1.
            stride: tp.Optional stride for the kernel. Either an integer or a sequence of
                length 1. Defaults to 1.
            padding: tp.Optional padding algorithm. Either ``VALID`` or ``SAME``.
                Defaults to ``SAME``. See:
                https://www.tensorflow.org/xla/operation_semantics#conv_convolution.
            with_bias: Whether to add a bias. By default, true.
            w_init: tp.Optional weight initialization. By default, truncated normal.
            b_init: tp.Optional bias initialization. By default, zeros.
            data_format: The data format of the input. Either ``NWC`` or ``NCW``. By
                default, ``NWC``.
            mask: tp.Optional mask of the weights.
            kwargs: Additional keyword arguments passed to Module.
        """
        super().__init__(
            num_spatial_dims=1,
            output_channels=output_channels,
            kernel_shape=kernel_shape,
            stride=stride,
            padding=padding,
            with_bias=with_bias,
            w_init=w_init,
            b_init=b_init,
            data_format=data_format,
            mask=mask,
            **kwargs,
        )


class Conv2DTranspose(ConvNDTranspose):
    """Two dimensional transposed convolution (aka. deconvolution)."""

    def __init__(
        self,
        output_channels: int,
        kernel_shape: tp.Union[int, tp.Sequence[int]],
        stride: tp.Union[int, tp.Sequence[int]] = 1,
        padding: tp.Union[str, tp.Sequence[tp.Tuple[int, int]]] = "SAME",
        with_bias: bool = True,
        w_init: tp.Optional[types.Initializer] = None,
        b_init: tp.Optional[types.Initializer] = None,
        data_format: str = "NHWC",
        mask: tp.Optional[np.ndarray] = None,
        **kwargs,
    ):
        """
        Initializes the module.

        Args:
            output_channels: Number of output channels.
            kernel_shape: The shape of the kernel. Either an integer or a sequence of
                length 2.
            stride: tp.Optional stride for the kernel. Either an integer or a sequence of
                length 2. Defaults to 1.
            padding: tp.Optional padding algorithm. Either ``VALID`` or ``SAME``.
                Defaults to ``SAME``. See:
                https://www.tensorflow.org/xla/operation_semantics#conv_convolution.
            with_bias: Whether to add a bias. By default, true.
            w_init: tp.Optional weight initialization. By default, truncated normal.
            b_init: tp.Optional bias initialization. By default, zeros.
            data_format: The data format of the input. Either ``NHWC`` or ``NCHW``. By
                default, ``NHWC``.
            mask: tp.Optional mask of the weights.
            kwargs: Additional keyword arguments passed to Module.
        """
        super().__init__(
            num_spatial_dims=2,
            output_channels=output_channels,
            kernel_shape=kernel_shape,
            stride=stride,
            padding=padding,
            with_bias=with_bias,
            w_init=w_init,
            b_init=b_init,
            data_format=data_format,
            mask=mask,
            **kwargs,
        )


class Conv3DTranspose(ConvNDTranspose):
    """Three dimensional transposed convolution (aka. deconvolution)."""

    def __init__(
        self,
        output_channels: int,
        kernel_shape: tp.Union[int, tp.Sequence[int]],
        stride: tp.Union[int, tp.Sequence[int]] = 1,
        padding: tp.Union[str, tp.Sequence[tp.Tuple[int, int]]] = "SAME",
        with_bias: bool = True,
        w_init: tp.Optional[types.Initializer] = None,
        b_init: tp.Optional[types.Initializer] = None,
        data_format: str = "NDHWC",
        mask: tp.Optional[np.ndarray] = None,
        **kwargs,
    ):
        """
        Initializes the module.

        Args:
            output_channels: Number of output channels.
            kernel_shape: The shape of the kernel. Either an integer or a sequence of
                length 3.
            stride: tp.Optional stride for the kernel. Either an integer or a sequence of
                length 3. Defaults to 1.
            padding: tp.Optional padding algorithm. Either ``VALID`` or ``SAME``.
                Defaults to ``SAME``. See:
                https://www.tensorflow.org/xla/operation_semantics#conv_convolution.
            with_bias: Whether to add a bias. By default, true.
            w_init: tp.Optional weight initialization. By default, truncated normal.
            b_init: tp.Optional bias initialization. By default, zeros.
            data_format: The data format of the input. Either ``NDHWC`` or ``NCDHW``.
                By default, ``NDHWC``.
            mask: tp.Optional mask of the weights.
            kwargs: Additional keyword arguments passed to Module.
        """
        super().__init__(
            num_spatial_dims=3,
            output_channels=output_channels,
            kernel_shape=kernel_shape,
            stride=stride,
            padding=padding,
            with_bias=with_bias,
            w_init=w_init,
            b_init=b_init,
            data_format=data_format,
            mask=mask,
            **kwargs,
        )



import functools
import typing as tp

import jax
import jax.numpy as jnp
import toolz
from elegy import hooks, types, utils

from .generalized_module import GeneralizedModule, register_module_for

try:
    import haiku
    from haiku._src.base import current_bundle_name, new_context
    from haiku._src.transform import (
        APPLY_RNG_ERROR,
        APPLY_RNG_STATE_ERROR,
        INIT_RNG_ERROR,
        check_mapping,
        to_prng_sequence,
    )
except ImportError:
    raise types.DependencyUnavailable("'haiku' is not available")


@register_module_for(haiku.Module)
class DummyHaikuModule(GeneralizedModule):
    pass


class TransformWithStateAndOutput(tp.NamedTuple):
    f: tp.Callable

    def init(
        self,
        rng: tp.Optional[tp.Union[jnp.ndarray, int]],
        *args,
        **kwargs,
    ) -> tp.Tuple[tp.Any, haiku.Params, haiku.State]:
        """Initializes your function collecting parameters and state."""
        rng = to_prng_sequence(rng, err_msg=INIT_RNG_ERROR)
        with new_context(rng=rng) as ctx:
            output = self.f(*args, **kwargs)
        return output, ctx.collect_params(), ctx.collect_initial_state()

    def apply(
        self,
        params: tp.Optional[haiku.Params],
        state: tp.Optional[haiku.State],
        rng: tp.Optional[tp.Union[jnp.ndarray, int]],
        *args,
        **kwargs,
    ) -> tp.Tuple[tp.Any, haiku.State]:
        """Applies your function injecting parameters and state."""
        params = check_mapping("params", params)
        state = check_mapping("state", state)
        rng = to_prng_sequence(
            rng, err_msg=(APPLY_RNG_STATE_ERROR if state else APPLY_RNG_ERROR)
        )
        with new_context(params=params, state=state, rng=rng) as ctx:
            out = self.f(*args, **kwargs)
        return out, ctx.collect_state()


class HaikuModule(GeneralizedModule):
    def __init__(self, f: tp.Callable):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            kwargs = {
                name[2:] if name.startswith("__") else name: value
                for name, value in kwargs.items()
            }
            return f(*args, **kwargs)

        self.f = wrapper
        self.module = TransformWithStateAndOutput(wrapper)

    def init(self, rng: types.RNGSeq) -> tp.Callable[..., types.OutputStates]:
        def _lambda(*args, **kwargs):
            def init_fn(*args, **kwargs) -> types.OutputStates:
                kwargs = {f"__{name}": value for name, value in kwargs.items()}
                key = rng.next()
                y_pred, params, states = self.module.init(key, *args, **kwargs)
                return types.OutputStates(y_pred, params, states)

            y_pred, params, states = utils.inject_dependencies(
                init_fn,
                signature_f=self.f,
            )(
                *args,
                **kwargs,
            )

            return types.OutputStates(y_pred, params, states)

        return _lambda

    def apply(
        self,
        params: tp.Any,
        states: tp.Any,
        training: bool,
        rng: types.RNGSeq,
    ) -> tp.Callable[..., types.OutputStates]:
        if params is None:
            params = {}

        if states is None:
            states = {}

        def _lambda(*args, **kwargs):
            def apply_fn(*args, **kwargs):
                kwargs = {f"__{name}": value for name, value in kwargs.items()}
                return self.module.apply(params, states, rng.next(), *args, **kwargs)

            y_pred, states_ = utils.inject_dependencies(apply_fn, signature_f=self.f,)(
                *args,
                **kwargs,
            )

            return types.OutputStates(y_pred, params, states_)

        return _lambda

    def get_summary_params(
        self,
        path: tp.Tuple[str, ...],
        module: tp.Any,
        value: tp.Any,
        net_params: types.NetParams,
        net_states: types.NetStates,
    ) -> tp.Tuple[tp.Optional[types.Pytree], tp.Optional[types.Pytree]]:

        path_str = "/".join(path)

        params_tree = net_params[path_str] if path_str in net_params else None
        states_tree = net_states[path_str] if path_str in net_states else None

        return params_tree, states_tree


def haiku_summarize(f):
    @functools.wraps(f)
    def wrapper(self: haiku.Module, *args, **kwargs):

        outputs = f(self, *args, **kwargs)

        if hooks.summaries_active():
            path = current_bundle_name().split("/")
            hooks.add_summary(tuple(path), self, outputs)

        return outputs

    return wrapper


def haiku_summary(
    name: str,
    f: tp.Any,
    value: types.Scalar,
):
    if hooks.summaries_active():
        path = tuple(current_bundle_name().split("/")) + (name,)
        hooks.add_summary(path, f, value)


def assert_id(value):
    assert value
    return value



from elegy import types
import typing as tp

import jax.numpy as jnp

from elegy import utils
from elegy.losses.loss import Loss, Reduction


def mean_squared_logarithmic_error(
    y_true: jnp.ndarray, y_pred: jnp.ndarray
) -> jnp.ndarray:
    """
    Computes the mean squared logarithmic error between labels and predictions.

    ```python
    loss = mean(square(log(y_true + 1) - log(y_pred + 1)), axis=-1)
    ```

    Usage:

    ```python
    rng = jax.random.PRNGKey(42)

    y_true = jax.random.randint(rng, shape=(2, 3), minval=0, maxval=2)
    y_pred = jax.random.uniform(rng, shape=(2, 3))

    loss = elegy.losses.mean_squared_logarithmic_error(y_true, y_pred)

    assert loss.shape == (2,)

    first_log = jnp.log(jnp.maximum(y_true, types.EPSILON) + 1.0)
    second_log = jnp.log(jnp.maximum(y_pred, types.EPSILON) + 1.0)
    assert jnp.array_equal(loss, jnp.mean(jnp.square(first_log - second_log), axis=-1))
    ```

    Arguments:
        y_true: Ground truth values. shape = `[batch_size, d0, .. dN]`.
        y_pred: The predicted values. shape = `[batch_size, d0, .. dN]`.

    Returns:
        Mean squared logarithmic error values. shape = `[batch_size, d0, .. dN-1]`.
    """

    y_true = y_true.astype(y_pred.dtype)
    first_log = jnp.log(jnp.maximum(y_true, types.EPSILON) + 1.0)
    second_log = jnp.log(jnp.maximum(y_pred, types.EPSILON) + 1.0)

    return jnp.mean(jnp.square(first_log - second_log), axis=-1)


class MeanSquaredLogarithmicError(Loss):
    """
    Computes the mean squared logarithmic errors between labels and predictions.

    `loss = mean(square(log(y_true + 1) - log(y_pred + 1)), axis=-1)`

    Usage:

    ```python
    y_true = jnp.array([[0.0, 1.0], [0.0, 0.0]])
    y_pred = jnp.array([[1.0, 1.0], [1.0, 0.0]])

    # Using 'auto'/'sum_over_batch_size' reduction type.
    msle = elegy.losses.MeanSquaredLogarithmicError()

    assert msle(y_true, y_pred) == 0.24022643

    # Calling with 'sample_weight'.
    assert msle(y_true, y_pred, sample_weight=jnp.array([0.7, 0.3])) = 0.12011322

    # Using 'sum' reduction type.
    msle = elegy.losses.MeanSquaredLogarithmicError(reduction=elegy.losses.Reduction.SUM)

    assert msle(y_true, y_pred) == 0.48045287

    # Using 'none' reduction type.
    msle = elegy.losses.MeanSquaredLogarithmicError(reduction=elegy.losses.Reduction.NONE)

    assert jnp.equal(msle(y_true, y_pred), jnp.array([0.24022643, 0.24022643])).all()
    ```
    Usage with the Elegy API:

    ```python
    model = elegy.Model(
        module_fn,
        loss=elegy.losses.MeanSquaredLogarithmicError(),
        metrics=elegy.metrics.Mean(),
    )
    ```
    """

    def __init__(
        self,
        reduction: tp.Optional[Reduction] = None,
        weight: tp.Optional[float] = None,
        on: tp.Optional[types.IndexLike] = None,
        **kwargs
    ):
        """
        Initializes `Mean` class.

        Arguments:
            reduction: (Optional) Type of `elegy.losses.Reduction` to apply to
                loss. Default value is `SUM_OVER_BATCH_SIZE`. For almost all cases
                this defaults to `SUM_OVER_BATCH_SIZE`.
            weight: Optional weight contribution for the total loss. Defaults to `1`.
            on: A string or integer, or iterable of string or integers, that
                indicate how to index/filter the `y_true` and `y_pred`
                arguments before passing them to `call`. For example if `on = "a"` then
                `y_true = y_true["a"]`. If `on` is an iterable
                the structures will be indexed iteratively, for example if `on = ["a", 0, "b"]`
                then `y_true = y_true["a"][0]["b"]`, same for `y_pred`. For more information
                check out [Keras-like behavior](https://poets-ai.github.io/elegy/guides/modules-losses-metrics/#keras-like-behavior).
        """

        return super().__init__(reduction=reduction, weight=weight, on=on, **kwargs)

    def call(
        self,
        y_true: jnp.ndarray,
        y_pred: jnp.ndarray,
        sample_weight: tp.Optional[
            jnp.ndarray
        ] = None,  # not used, __call__ handles it, left for documentation purposes.
    ) -> jnp.ndarray:
        """
        Invokes the `MeanSquaredLogarithmicError` instance.

        Arguments:
            y_true: Ground truth values. shape = `[batch_size, d0, .. dN]`, except
                sparse loss functions such as sparse categorical crossentropy where
                shape = `[batch_size, d0, .. dN-1]`
            y_pred: The predicted values. shape = `[batch_size, d0, .. dN]`
            sample_weight: Optional `sample_weight` acts as a
                coefficient for the loss. If a scalar is provided, then the loss is
                simply scaled by the given value. If `sample_weight` is a tensor of size
                `[batch_size]`, then the total loss for each sample of the batch is
                rescaled by the corresponding element in the `sample_weight` vector. If
                the shape of `sample_weight` is `[batch_size, d0, .. dN-1]` (or can be
                broadcasted to this shape), then each loss element of `y_pred` is scaled
                by the corresponding value of `sample_weight`. (Note on`dN-1`: all loss
                functions reduce by 1 dimension, usually axis=-1.)

        Returns:
            Weighted loss float `Tensor`. If `reduction` is `NONE`, this has
                shape `[batch_size, d0, .. dN-1]`; otherwise, it is scalar. (Note `dN-1`
                because all loss functions reduce by 1 dimension, usually axis=-1.)

        Raises:
            ValueError: If the shape of `sample_weight` is invalid.
        """
        return mean_squared_logarithmic_error(y_true, y_pred)



import jax.numpy as jnp

from unittest import TestCase
import jax

import elegy


class SequentialTest(TestCase):
    #
    def test_connects(self):

        y = elegy.nn.Sequential(
            lambda: [
                elegy.nn.Flatten(),
                elegy.nn.Linear(5),
                jax.nn.relu,
                elegy.nn.Linear(2),
            ]
        ).call_with_defaults(rng=elegy.RNGSeq(42))(jnp.ones([10, 3]))

        assert y.shape == (10, 2)

        y = elegy.nn.Sequential(
            lambda: [
                elegy.nn.Flatten(),
                elegy.nn.Linear(5),
                jax.nn.relu,
                elegy.nn.Linear(2),
            ]
        ).call_with_defaults(rng=elegy.RNGSeq(42), training=False)(jnp.ones([10, 3]))

        assert y.shape == (10, 2)

    def test_di(self):

        m = elegy.nn.Sequential(
            lambda: [
                elegy.nn.Flatten(),
                elegy.nn.Linear(2),
            ]
        )

        y = elegy.inject_dependencies(
            m.call_with_defaults(rng=elegy.RNGSeq(42), training=False), signature_f=m
        )(
            jnp.ones([5, 3]),
            a=1,
            b=2,
        )

        assert y.shape == (5, 2)



import typing as tp

import haiku as hk
import jax.numpy as jnp
import numpy as np
from elegy import hooks, module, types
from elegy.initializers import TruncatedNormal


class Linear(module.Module):
    """Linear module."""

    w: np.ndarray
    b: np.ndarray

    def __init__(
        self,
        output_size: int,
        with_bias: bool = True,
        w_init: tp.Optional[types.Initializer] = None,
        b_init: tp.Optional[types.Initializer] = None,
        **kwargs
    ):
        """
        Constructs the Linear module.

        Arguments:
            output_size: Output dimensionality.
            with_bias: Whether to add a bias to the output.
            w_init: Optional initializer for weights. By default, uses random values
                from truncated normal, with stddev `1 / sqrt(fan_in)`. See
                https://arxiv.org/abs/1502.03167v3.
            b_init: Optional initializer for bias. By default, zero.
            kwargs: Additional keyword arguments passed to Module.
        """
        super().__init__(**kwargs)
        self.input_size = None
        self.output_size = output_size
        self.with_bias = with_bias
        self.w_init = w_init
        self.b_init = b_init or jnp.zeros

    def call(self, inputs: np.ndarray) -> np.ndarray:
        """"""
        if not inputs.shape:
            raise ValueError("Input must not be scalar.")

        input_size = self.input_size = inputs.shape[-1]
        output_size = self.output_size
        dtype = jnp.float32

        w_init = self.w_init

        if w_init is None:
            stddev = 1.0 / np.sqrt(self.input_size)
            w_init = TruncatedNormal(stddev=stddev)

        w = self.add_parameter("w", lambda: w_init([input_size, output_size], dtype))

        inputs = jnp.asarray(inputs, dtype=self.dtype)
        w = jnp.asarray(w, self.dtype)
        out = jnp.dot(inputs, w)

        if self.with_bias:
            b = self.add_parameter("b", lambda: self.b_init([self.output_size], dtype))
            b = jnp.expand_dims(b, axis=tuple(range(len(out.shape) - 1)))
            b = jnp.broadcast_to(b, out.shape)
            b = jnp.asarray(b, self.dtype)
            out = out + b

        return out



# Implementation based on tf.keras.callbacks.py
# https://github.com/tensorflow/tensorflow/blob/v2.2.0/tensorflow/python/keras/callbacks.py


import os
import typing as tp
from typing import Any, Dict, Optional, Union
from tensorboardX.writer import SummaryWriter

from .callback import Callback


class TensorBoard(Callback):
    """
    Callback that streams epoch results to tensorboard events folder.

    Supports all values that can be represented as a string,
    including 1D iterables such as `np.ndarray`.


    ```python
    tensorboard_logger = TensorBoard('runs')
    model.fit(X_train, Y_train, callbacks=[tensorboard_logger])
    ```
    """

    def __init__(
        self,
        logdir: Optional[str] = None,
        *,
        update_freq: Union[str, int] = "epoch",
        purge_step: Optional[int] = None,
        comment: str = "",
    ) -> None:
        """
        Arguments:
            logdir: Save directory location. Default is
                runs/**CURRENT_DATETIME_HOSTNAME**/{train, val}, which changes after each run.
                Use hierarchical folder structure to compare
                between runs easily. e.g. pass in 'runs/exp1', 'runs/exp2', etc.
                for each new experiment to compare across them.
            update_freq: `'batch'` or `'epoch'` or integer. When using `'batch'`,
                writes the losses and metrics to TensorBoard after each batch. The same
                applies for `'epoch'`. If using an integer, let's say `1000`, the
                callback will write the metrics and losses to TensorBoard every 1000
                batches. Note that writing too frequently to TensorBoard can slow down
                your training.
            purge_step (int):
                When logging crashes at step :math:`T+X` and restarts at step :math:`T`,
                any events whose global_step larger or equal to :math:`T` will be
                purged and hidden from TensorBoard.
                Note that crashed and resumed experiments should have the same ``logdir``.
            comment (string): Comment logdir suffix appended to the default
                ``logdir``. If ``logdir`` is assigned, this argument has no effect.
        """
        if not logdir:
            import socket
            from datetime import datetime

            current_time = datetime.now().strftime("%b%d_%H-%M-%S")
            self.logdir = os.path.join(
                "runs", current_time + "_" + socket.gethostname() + comment
            )
        else:
            self.logdir = logdir
        self.train_writer = None
        self.val_writer = None
        self.keys = None
        self.write_per_batch = True
        try:
            self.update_freq = int(update_freq)
        except ValueError as e:
            self.update_freq = 1
            if update_freq == "batch":
                self.write_per_batch = True
            elif update_freq == "epoch":
                self.write_per_batch = False
            else:
                raise e
        self.purge_step = purge_step

        super(TensorBoard, self).__init__()

    def on_train_begin(self, logs=None):
        self.train_writer = SummaryWriter(
            os.path.join(self.logdir, "train"), purge_step=self.purge_step
        )
        self.val_writer = SummaryWriter(
            os.path.join(self.logdir, "val"), purge_step=self.purge_step
        )
        self.steps = self.params["steps"]
        self.global_step = 0

    def on_train_batch_end(self, batch: int, logs=None):
        if not self.write_per_batch:
            return
        logs = logs or {}
        self.global_step = batch + self.current_epoch * (self.steps)
        if self.global_step % self.update_freq == 0:
            if self.keys is None:
                self.keys = logs.keys()
            for key in self.keys:
                self.train_writer.add_scalar(key, logs[key], self.global_step)

    def on_epoch_begin(self, epoch: int, logs=None):
        self.current_epoch = epoch

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}

        if self.keys is None:
            self.keys = logs.keys()

        # logs on on_{train, test}_batch_end do not have val metrics
        if self.write_per_batch:
            for key in logs:
                if "val" in key:
                    self.val_writer.add_scalar(
                        key.replace("val_", ""), logs[key], self.global_step
                    )
            return

        elif epoch % self.update_freq == 0:

            for key in self.keys:
                if "val" in key:
                    self.val_writer.add_scalar(
                        key.replace("val_", ""), logs[key], epoch
                    )
                else:
                    self.train_writer.add_scalar(key, logs[key], epoch)

    def on_train_end(self, logs=None):
        self.train_writer.close()
        self.val_writer.close()



from elegy import utils

import jax.numpy as jnp
from unittest import TestCase

import elegy


class DropoutTest(TestCase):
    def test_dropout_connects(self):
        elegy.nn.Dropout(0.25).call_with_defaults(rng=elegy.RNGSeq(42))(
            jnp.ones([3, 3]), training=True
        )

    def test_on_predict(self):
        class TestModule(elegy.Module):
            def call(self, x, training):
                return elegy.nn.Dropout(0.5)(x, training)

        model = elegy.Model(TestModule())

        x = jnp.ones([3, 5])

        y_pred = model.predict(x, initialize=True)
        logs = model.evaluate(x)

        assert jnp.all(y_pred == x)



from elegy import types
import typing as tp

import jax
import jax.numpy as jnp

from elegy import utils, types
from elegy.losses.loss import Loss, Reduction
from elegy.losses.categorical_crossentropy import categorical_crossentropy


def sparse_categorical_crossentropy(
    y_true: jnp.ndarray,
    y_pred: jnp.ndarray,
    from_logits: bool = False,
    check_bounds: bool = True,
) -> jnp.ndarray:

    n_classes = y_pred.shape[-1]

    if from_logits:
        y_pred = jax.nn.log_softmax(y_pred)
        loss = -jnp.take_along_axis(y_pred, y_true[..., None], axis=-1)[..., 0]
    else:
        # select output value
        y_pred = jnp.take_along_axis(y_pred, y_true[..., None], axis=-1)[..., 0]

        # calculate log
        y_pred = jnp.maximum(y_pred, types.EPSILON)
        y_pred = jnp.log(y_pred)
        loss = -y_pred

    if check_bounds:
        # set NaN where y_true is negative or larger/equal to the number of y_pred channels
        loss = jnp.where(y_true < 0, jnp.nan, loss)
        loss = jnp.where(y_true >= n_classes, jnp.nan, loss)

    return loss


class SparseCategoricalCrossentropy(Loss):
    """
    Computes the crossentropy loss between the labels and predictions.

    Use this crossentropy loss function when there are two or more label classes.
    We expect labels to be provided as integers. If you want to provide labels
    using `one-hot` representation, please use `CategoricalCrossentropy` loss.
    There should be `# classes` floating point values per feature for `y_pred`
    and a single floating point value per feature for `y_true`.
    In the snippet below, there is a single floating point value per example for
    `y_true` and `# classes` floating pointing values per example for `y_pred`.
    The shape of `y_true` is `[batch_size]` and the shape of `y_pred` is
    `[batch_size, num_classes]`.

    Usage:
    ```python
    y_true = jnp.array([1, 2])
    y_pred = jnp.array([[0.05, 0.95, 0], [0.1, 0.8, 0.1]])

    # Using 'auto'/'sum_over_batch_size' reduction type.
    scce = elegy.losses.SparseCategoricalCrossentropy()
    result = scce(y_true, y_pred)  # 1.177
    assert jnp.isclose(result, 1.177, rtol=0.01)

    # Calling with 'sample_weight'.
    result = scce(y_true, y_pred, sample_weight=jnp.array([0.3, 0.7]))  # 0.814
    assert jnp.isclose(result, 0.814, rtol=0.01)

    # Using 'sum' reduction type.
    scce = elegy.losses.SparseCategoricalCrossentropy(
        reduction=elegy.losses.Reduction.SUM
    )
    result = scce(y_true, y_pred)  # 2.354
    assert jnp.isclose(result, 2.354, rtol=0.01)

    # Using 'none' reduction type.
    scce = elegy.losses.SparseCategoricalCrossentropy(
        reduction=elegy.losses.Reduction.NONE
    )
    result = scce(y_true, y_pred)  # [0.0513, 2.303]
    assert jnp.all(jnp.isclose(result, [0.0513, 2.303], rtol=0.01))
    ```

    Usage with the `Elegy` API:

    ```python
    model = elegy.Model(
        module_fn,
        loss=elegy.losses.SparseCategoricalCrossentropy(),
        metrics=elegy.metrics.Accuracy(),
        optimizer=optax.adam(1e-3),
    )

    ```
    """

    def __init__(
        self,
        from_logits: bool = False,
        reduction: tp.Optional[Reduction] = None,
        weight: tp.Optional[float] = None,
        on: tp.Optional[types.IndexLike] = None,
        check_bounds: tp.Optional[bool] = True,
        **kwargs
    ):
        """
        Initializes `SparseCategoricalCrossentropy` instance.

        Arguments:
            from_logits: Whether `y_pred` is expected to be a logits tensor. By
                default, we assume that `y_pred` encodes a probability distribution.
                **Note - Using from_logits=True is more numerically stable.**
            reduction: (Optional) Type of `elegy.losses.Reduction` to apply to
                loss. Default value is `SUM_OVER_BATCH_SIZE`. For almost all cases
                this defaults to `SUM_OVER_BATCH_SIZE`.
            weight: Optional weight contribution for the total loss. Defaults to `1`.
            on: A string or integer, or iterable of string or integers, that
                indicate how to index/filter the `y_true` and `y_pred`
                arguments before passing them to `call`. For example if `on = "a"` then
                `y_true = y_true["a"]`. If `on` is an iterable
                the structures will be indexed iteratively, for example if `on = ["a", 0, "b"]`
                then `y_true = y_true["a"][0]["b"]`, same for `y_pred`. For more information
                check out [Keras-like behavior](https://poets-ai.github.io/elegy/guides/modules-losses-metrics/#keras-like-behavior).
            check_bounds: If `True` (default), checks `y_true` for negative values and values
                larger or equal than the number of channels in `y_pred`. Sets loss to NaN
                if this is the case. If `False`, the check is disabled and the loss may contain
                incorrect values.
        """
        super().__init__(reduction=reduction, weight=weight, on=on, **kwargs)

        self._from_logits = from_logits
        self._check_bounds = check_bounds

    def call(
        self, y_true, y_pred, sample_weight: tp.Optional[jnp.ndarray] = None
    ) -> jnp.ndarray:
        """
        Invokes the `SparseCategoricalCrossentropy` instance.

        Arguments:
            y_true: Ground truth values.
            y_pred: The predicted values.
            sample_weight: Acts as a
                coefficient for the loss. If a scalar is provided, then the loss is
                simply scaled by the given value. If `sample_weight` is a tensor of size
                `[batch_size]`, then the total loss for each sample of the batch is
                rescaled by the corresponding element in the `sample_weight` vector. If
                the shape of `sample_weight` is `[batch_size, d0, .. dN-1]` (or can be
                broadcasted to this shape), then each loss element of `y_pred` is scaled
                by the corresponding value of `sample_weight`. (Note on`dN-1`: all loss
                functions reduce by 1 dimension, usually axis=-1.)

        Returns:
            Loss values per sample.
        """

        return sparse_categorical_crossentropy(
            y_true,
            y_pred,
            from_logits=self._from_logits,
            check_bounds=self._check_bounds,
        )



import jax
import jax.numpy as jnp
import unittest
import pytest

import elegy


class TestHooks(unittest.TestCase):
    def test_losses(self):
        assert not elegy.hooks.losses_active()

        with elegy.hooks.context(set_all=True):
            elegy.hooks.add_loss("x", 2.0)
            losses = elegy.hooks.get_losses()

        assert losses["x_loss"] == 2.0

    def test_metrics(self):
        assert not elegy.hooks.metrics_active()

        with elegy.hooks.context(set_all=True):
            elegy.hooks.add_metric("x", 2.0)
            metrics = elegy.hooks.get_metrics()

        assert metrics["x"] == 2.0

    def test_summaries(self):
        assert not elegy.hooks.summaries_active()

        with elegy.hooks.context(summaries=True):
            elegy.hooks.add_summary(("a", 0, "b"), None, 2.0)
            summaries = elegy.hooks.get_summaries()

        assert summaries[0] == (("a", 0, "b"), None, 2.0)

    def test_no_summaries(self):
        assert not elegy.hooks.summaries_active()

        with elegy.hooks.context(summaries=False):
            elegy.hooks.add_summary(("a", 0, "b"), None, 2.0)
            has_summaries = elegy.hooks.summaries_active()

        assert not has_summaries

    def test_jit(self):
        assert not elegy.hooks.losses_active()

        def f(x):
            x = 2.0 * x
            elegy.hooks.add_loss("x", x)
            elegy.hooks.add_metric("x", x + 1)
            elegy.hooks.add_summary(("a", 0, "b"), jax.nn.relu, x + 2)

            return x

        f_ = elegy.hooks.jit(f)

        with elegy.hooks.context(set_all=True):
            x = f_(3.0)
            losses = elegy.hooks.get_losses()
            metrics = elegy.hooks.get_metrics()
            summaries = elegy.hooks.get_summaries()

        assert x == 6
        assert losses["x_loss"] == 6
        assert metrics["x"] == 7
        assert summaries[0] == (("a", 0, "b"), jax.nn.relu, 8)



from unittest import TestCase

import jax.numpy as jnp
import tensorflow_addons as tfa
import numpy as np
import elegy


class F1Test(TestCase):
    #
    def test_basic(self):

        y_true = jnp.array([0, 1, 1, 1])
        y_pred = jnp.array([1, 0, 1, 1])

        assert np.allclose(
            elegy.metrics.F1().call_with_defaults()(y_true, y_pred),
            tfa.metrics.F1Score(2, average="micro", threshold=0.5)(y_true, y_pred),
        )  # 2 * (0.44445 / 1.33334)

        y_true = jnp.array([1, 1, 1, 1])
        y_pred = jnp.array([1, 1, 0, 0])

        assert np.allclose(
            elegy.metrics.F1().call_with_defaults()(y_true, y_pred),
            tfa.metrics.F1Score(2, average="micro", threshold=0.5)(y_true, y_pred),
        )  # 2 * (0.5 / 1.5)

        y_true = (np.random.uniform(0, 1, size=(5, 6, 7)) > 0.5).astype(np.float32)
        y_pred = np.random.uniform(0, 1, size=(5, 6, 7))
        sample_weight = np.expand_dims(
            (np.random.uniform(0, 1, size=(6, 7)) > 0.5).astype(np.float32), axis=0
        )

        assert np.allclose(
            tfa.metrics.F1Score(2, average="micro", threshold=0.3)(y_true, y_pred),
            elegy.metrics.F1(threshold=0.3).call_with_defaults()(
                jnp.asarray(y_true), jnp.asarray(y_pred)
            ),
        )

        assert np.allclose(
            tfa.metrics.F1Score(2, average="micro", threshold=0.3)(
                y_true, y_pred, sample_weight=sample_weight
            ),
            elegy.metrics.F1(threshold=0.3).call_with_defaults()(
                jnp.asarray(y_true), jnp.asarray(y_pred), sample_weight=sample_weight
            ),
        )

    #
    def test_cumulative(self):
        em = elegy.metrics.F1(threshold=0.3)
        tm = tfa.metrics.F1Score(2, average="micro", threshold=0.3)

        # 1st run
        y_true = (np.random.uniform(0, 1, size=(5, 6, 7)) > 0.5).astype(np.float32)
        y_pred = np.random.uniform(0, 1, size=(5, 6, 7))
        sample_weight = np.expand_dims(
            (np.random.uniform(0, 1, size=(6, 7)) > 0.5).astype(np.float32), axis=0
        )

        assert np.allclose(
            tm(y_true, y_pred, sample_weight=sample_weight),
            em.call_with_defaults()(
                jnp.asarray(y_true),
                jnp.asarray(y_pred),
                sample_weight=jnp.asarray(sample_weight),
            ),
        )

        # 2nd run
        y_true = (np.random.uniform(0, 1, size=(5, 6, 7)) > 0.5).astype(np.float32)
        y_pred = np.random.uniform(0, 1, size=(5, 6, 7))
        sample_weight = np.expand_dims(
            (np.random.uniform(0, 1, size=(6, 7)) > 0.5).astype(np.float32), axis=0
        )

        assert np.allclose(
            tm(y_true, y_pred, sample_weight=sample_weight),
            em.call_with_defaults()(
                jnp.asarray(y_true),
                jnp.asarray(y_pred),
                sample_weight=jnp.asarray(sample_weight),
            ),
        )



import unittest
from hashlib import new
from tempfile import TemporaryDirectory
from pathlib import Path

import cloudpickle
import elegy
import jax
import jax.numpy as jnp
import numpy as np
import optax
import pytest
import sh
import tensorflow as tf


class MLP(elegy.Module):
    """Standard LeNet-300-100 MLP network."""

    n1: int
    n2: int

    def __init__(self, n1: int = 3, n2: int = 4):
        super().__init__()
        self.n1 = n1
        self.n2 = n2

    def call(self, image: jnp.ndarray, training: bool):
        x = image.astype(jnp.float32) / 255.0

        x = jnp.reshape(x, [x.shape[0], -1])
        x = elegy.nn.Linear(self.n1)(x)
        x = elegy.nn.BatchNormalization()(x)
        x = jax.nn.relu(x)

        x = elegy.nn.Linear(self.n2)(x)
        x = jax.nn.relu(x)
        x = elegy.nn.Linear(10)(x)

        return x


class ModelBasicTest(unittest.TestCase):
    def test_predict(self):

        model = elegy.Model(module=elegy.nn.Linear(1))

        X = np.random.uniform(size=(5, 10))
        y = np.random.randint(10, size=(5, 1))

        model.init(x=X, y=y)
        y_pred = model.predict(x=X)

        assert y_pred.shape == (5, 1)

    def test_evaluate(self):
        def mse(y_true, y_pred):
            return jnp.mean((y_true - y_pred) ** 2)

        def mae(y_true, y_pred):
            return jnp.mean(jnp.abs(y_true - y_pred))

        model = elegy.Model(
            module=elegy.nn.Linear(1),
            loss=dict(a=mse),
            metrics=dict(b=mae),
            optimizer=optax.adamw(1e-3),
            run_eagerly=True,
        )

        X = np.random.uniform(size=(5, 10))
        y = np.random.uniform(size=(5, 1))

        logs = model.evaluate(x=X, y=y)

        assert "a/mse_loss" in logs
        assert "b/mae" in logs
        assert "loss" in logs

    def test_metrics(self):
        class M(elegy.Module):
            def call(self, x):
                n = self.add_parameter("n", lambda: 0, trainable=False)
                self.update_parameter("n", n + 1)
                return x

        metrics = elegy.model.model.Metrics(dict(a=dict(b=[M(), M()], c=M())))

        rng = elegy.RNGSeq(42)
        x = np.random.uniform(size=(5, 7, 7))

        with elegy.hooks.context(metrics=True):
            elegy.hooks.add_metric("d", 10)
            aux_metrics = elegy.hooks.get_metrics()
            logs, states = metrics.init(aux_metrics, rng)(x, training=True)

        with elegy.hooks.context(metrics=True):
            elegy.hooks.add_metric("d", 10)
            aux_metrics = elegy.hooks.get_metrics()
            logs, states = metrics.apply(aux_metrics, rng, states)(x, training=True)

        assert len(metrics.metrics) == 3
        assert "a/b/m" in metrics.metrics
        assert "a/b/m_1" in metrics.metrics
        assert "a/c/m" in metrics.metrics

        assert len(logs) == 4
        assert "a/b/m" in logs
        assert "a/b/m_1" in logs
        assert "a/c/m" in logs
        assert "d" in logs

        assert len(states) == 3
        assert "a/b/m" in states
        assert "a/b/m_1" in states
        assert "a/c/m" in states

    def test_losses(self):
        def loss_fn():
            return 3.0

        losses = elegy.model.model.Losses(dict(a=dict(b=[loss_fn, loss_fn], c=loss_fn)))

        rng = elegy.RNGSeq(42)
        hooks_losses = dict(x=0.3, y=4.5)

        with elegy.hooks.context(losses=True):
            elegy.hooks.add_loss("d", 1.0)
            aux_losses = elegy.hooks.get_losses()
            logs, logs, states = losses.init(aux_losses, rng)()

        with elegy.hooks.context(losses=True):
            elegy.hooks.add_loss("d", 1.0)
            aux_losses = elegy.hooks.get_losses()
            loss, logs, states = losses.apply(aux_losses, states)()

        assert loss == 10

        assert len(losses.losses) == 3
        assert "a/b/loss_fn" in losses.losses
        assert "a/b/loss_fn_1" in losses.losses
        assert "a/c/loss_fn" in losses.losses

        assert len(logs) == 5
        assert "loss" in logs
        assert "a/b/loss_fn" in logs
        assert "a/b/loss_fn_1" in logs
        assert "a/c/loss_fn" in logs
        assert "d_loss" in logs


class ModelTest(unittest.TestCase):
    def test_evaluate(self):

        model = elegy.Model(
            module=MLP(n1=3, n2=1),
            loss=[
                elegy.losses.SparseCategoricalCrossentropy(from_logits=True),
                elegy.regularizers.GlobalL2(l=1e-4),
            ],
            metrics=elegy.metrics.SparseCategoricalAccuracy(),
            optimizer=optax.adamw(1e-3),
            run_eagerly=True,
        )

        X = np.random.uniform(size=(5, 7, 7))
        y = np.random.randint(10, size=(5,))

        history = model.fit(
            x=X,
            y=y,
            epochs=1,
            steps_per_epoch=1,
            batch_size=5,
            validation_data=(X, y),
            shuffle=True,
            verbose=1,
        )

        logs = model.evaluate(X, y)

        eval_acc = logs["sparse_categorical_accuracy"]
        predict_acc = (model.predict(X).argmax(-1) == y).mean()

        assert eval_acc == predict_acc

    def test_cloudpickle(self):
        model = elegy.Model(
            module=MLP(n1=3, n2=1),
            loss=[
                elegy.losses.SparseCategoricalCrossentropy(from_logits=True),
                elegy.regularizers.GlobalL2(l=1e-4),
            ],
            metrics=elegy.metrics.SparseCategoricalAccuracy(),
            optimizer=optax.adamw(1e-3),
            run_eagerly=True,
        )

        X = np.random.uniform(size=(5, 7, 7))
        y = np.random.randint(10, size=(5,))

        model.init(X, y)
        y0 = model.predict(X)

        model_pkl = cloudpickle.dumps(model)
        newmodel = cloudpickle.loads(model_pkl)

        newmodel.states = model.states
        newmodel.initial_states = model.initial_states

        y1 = newmodel.predict(X)
        assert np.all(y0 == y1)

    def test_saved_model(self):

        with TemporaryDirectory() as model_dir:

            model = elegy.Model(module=elegy.nn.Linear(4))

            x = np.random.uniform(size=(5, 6))

            with pytest.raises(elegy.types.ModelNotInitialized):
                model.saved_model(x, model_dir, batch_size=[1, 2, 4, 8])

            model.init(x)
            model.saved_model(x, model_dir, batch_size=[1, 2, 4, 8])

            output = str(sh.ls(model_dir))

            assert "saved_model.pb" in output
            assert "variables" in output

            saved_model = tf.saved_model.load(model_dir)

            saved_model



import typing as tp

try:
    from jax.experimental import jax2tf  # type: ignore[import]
    import tensorflow as tf  # type: ignore[import]

    def convert_and_save_model(
        jax_fn: tp.Callable[[tp.Any, tp.Any], tp.Any],
        params,
        model_dir: str,
        *,
        input_signatures: tp.Sequence[tf.TensorSpec],
        shape_polymorphic_input_spec: tp.Optional[str] = None,
        with_gradient: bool = False,
        enable_xla: bool = True,
        compile_model: bool = True,
        save_model_options: tp.Optional[tf.saved_model.SaveOptions] = None
    ):
        """Convert a JAX function and saves a SavedModel.
        This is an example, for serious uses you will likely want to copy and
        expand it as needed (see note at the top of the model).
        Use this function if you have a trained ML model that has both a prediction
        function and trained parameters, which you want to save separately from the
        function graph as variables (e.g., to avoid limits on the size of the
        GraphDef, or to enable fine-tuning.) If you don't have such parameters,
        you can still use this library function but probably don't need it
        (see jax2tf/README.md for some simple examples).
        In order to use this wrapper you must first convert your model to a function
        with two arguments: the parameters and the input on which you want to do
        inference. Both arguments may be np.ndarray or (nested)
        tuples/lists/dictionaries thereof.
        See the README.md for a discussion of how to prepare Flax and Haiku models.
        Args:
        jax_fn: a JAX function taking two arguments, the parameters and the inputs.
            Both arguments may be (nested) tuples/lists/dictionaries of np.ndarray.
        params: the parameters, to be used as first argument for `jax_fn`. These
            must be (nested) tuples/lists/dictionaries of np.ndarray, and will be
            saved as the variables of the SavedModel.
        model_dir: the directory where the model should be saved.
        input_signatures: the input signatures for the second argument of `jax_fn`
            (the input). A signature must be a `tensorflow.TensorSpec` instance, or a
            (nested) tuple/list/dictionary thereof with a structure matching the
            second argument of `jax_fn`. The first input_signature will be saved as
            the default serving signature. The additional signatures will be used
            only to ensure that the `jax_fn` is traced and converted to TF for the
            corresponding input shapes.
        shape_polymorphic_input_spec: if given then it will be used as the
            `in_shapes` argument to jax2tf.convert for the second parameter of
            `jax_fn`. In this case, a single `input_signatures` is supported, and
            should have `None` in the polymorphic dimensions. Should be a string, or a
            (nesteD) tuple/list/dictionary thereof with a structure matching the
            second argument of `jax_fn`.
        with_gradient: whether the SavedModel should support gradients. If True,
            then a custom gradient is saved. If False, then a
            tf.raw_ops.PreventGradient is saved to error if a gradient is attempted.
            (At the moment due to a bug in SavedModel, custom gradients are not
            supported.)
        enable_xla: whether the jax2tf converter is allowed to use TFXLA ops. If
            False, the conversion tries harder to use purely TF ops and raises an
            exception if it is not possible. (default: True)
        compile_model: use TensorFlow jit_compiler on the SavedModel. This
            is needed if the SavedModel will be used for TensorFlow serving.
        save_model_options: options to pass to savedmodel.save.
        """

        if not input_signatures:
            raise ValueError("At least one input_signature must be given")
        if shape_polymorphic_input_spec is not None:
            if len(input_signatures) > 1:
                raise ValueError(
                    "For shape-polymorphic conversion a single "
                    "input_signature is supported."
                )
        tf_fn = jax2tf.convert(
            jax_fn,
            with_gradient=with_gradient,
            in_shapes=[None, shape_polymorphic_input_spec],
            enable_xla=enable_xla,
        )

        # Create tf.Variables for the parameters. If you want more useful variable
        # names, you can use `tree.map_structure_with_path` from the `dm-tree` package
        param_vars = tf.nest.map_structure(
            # Due to a bug in SavedModel it is not possible to use tf.GradientTape on
            # a function converted with jax2tf and loaded from SavedModel. Thus, we
            # mark the variables as non-trainable to ensure that users of the
            # SavedModel will not try to fine tune them.
            lambda param: tf.Variable(param, trainable=with_gradient),
            params,
        )
        tf_fun = tf.function(
            lambda inputs: tf_fn(param_vars, inputs),
            autograph=False,
            experimental_compile=compile_model,
        )

        signatures = {}
        # This signature is needed for TensorFlow Serving use.
        signatures[
            tf.saved_model.DEFAULT_SERVING_SIGNATURE_DEF_KEY
        ] = tf_fun.get_concrete_function(input_signatures[0])

        for input_signature in input_signatures[1:]:
            # If there are more signatures, trace and cache a TF function for each one
            tf_fun.get_concrete_function(input_signature)

        wrapper = _ReusableSavedModelWrapper(tf_fun, param_vars)
        tf.saved_model.save(
            wrapper, model_dir, signatures=signatures, options=save_model_options
        )

    class _ReusableSavedModelWrapper(tf.train.Checkpoint):
        """Wraps a function and its parameters for saving to a SavedModel.
        Implements the interface described at
        https://www.tensorflow.org/hub/reusable_saved_models.
        """

        def __init__(self, tf_fun, param_vars):
            """Args:
            tf_fun: a tf.function taking one argument (the inputs), which can be
            be tuples/lists/dictionaries of np.ndarray or tensors. The function
            may have references to the tf.Variables in `param_vars`.
            param_vars: the parameters, as tuples/lists/dictionaries of tf.Variable,
            to be saved as the variables of the SavedModel.
            """
            super().__init__()
            # Implement the interface from https://www.tensorflow.org/hub/reusable_saved_models
            self.variables = tf.nest.flatten(param_vars)
            self.trainable_variables = [v for v in self.variables if v.trainable]
            # If you intend to prescribe regularization terms for users of the model,
            # add them as @tf.functions with no inputs to this list. Else drop this.
            self.regularization_losses = []
            self.__call__ = tf_fun


except ImportError:
    convert_and_save_model = None



import jax
import jax.numpy as jnp
import tensorflow.keras as tfk
from unittest import TestCase

import elegy
from elegy import utils, types


class MeanAbsolutePercentageErrorTest(TestCase):
    #
    def test_basic(self):
        y_true = jnp.array([[1.0, 1.0], [0.9, 0.0]])
        y_pred = jnp.array([[1.0, 1.0], [1.0, 0.0]])

        # Using 'auto'/'sum_over_batch_size' reduction type.
        mape = elegy.losses.MeanAbsolutePercentageError()
        result = mape(y_true, y_pred)
        assert jnp.isclose(result, 2.78, rtol=0.01)

        # Calling with 'sample_weight'.
        assert jnp.isclose(
            mape(y_true, y_pred, sample_weight=jnp.array([0.1, 0.9])), 2.5, rtol=0.01
        )

        # Using 'sum' reduction type.
        mape = elegy.losses.MeanAbsolutePercentageError(
            reduction=elegy.losses.Reduction.SUM
        )

        assert jnp.isclose(mape(y_true, y_pred), 5.6, rtol=0.01)

        # Using 'none' reduction type.
        mape = elegy.losses.MeanAbsolutePercentageError(
            reduction=elegy.losses.Reduction.NONE
        )

        result = mape(y_true, y_pred)
        assert jnp.all(jnp.isclose(result, [0.0, 5.6], rtol=0.01))

    #
    def test_function(self):

        y_true = jnp.array([[1.0, 1.0], [0.9, 0.0]])
        y_pred = jnp.array([[1.0, 1.0], [1.0, 0.0]])

        ## Standard MAPE
        mape_elegy = elegy.losses.MeanAbsolutePercentageError()
        mape_tfk = tfk.losses.MeanAbsolutePercentageError()
        assert jnp.isclose(
            mape_elegy(y_true, y_pred), mape_tfk(y_true, y_pred), rtol=0.0001
        )

        ## MAPE using sample_weight
        assert jnp.isclose(
            mape_elegy(y_true, y_pred, sample_weight=jnp.array([1, 0])),
            mape_tfk(y_true, y_pred, sample_weight=jnp.array([1, 0])),
            rtol=0.0001,
        )

        ## MAPE with reduction method: SUM
        mape_elegy = elegy.losses.MeanAbsolutePercentageError(
            reduction=elegy.losses.Reduction.SUM
        )
        mape_tfk = tfk.losses.MeanAbsolutePercentageError(
            reduction=tfk.losses.Reduction.SUM
        )
        assert jnp.isclose(
            mape_elegy(y_true, y_pred), mape_tfk(y_true, y_pred), rtol=0.0001
        )

        ## MAPE with reduction method: NONE
        mape_elegy = elegy.losses.MeanAbsolutePercentageError(
            reduction=elegy.losses.Reduction.NONE
        )
        mape_tfk = tfk.losses.MeanAbsolutePercentageError(
            reduction=tfk.losses.Reduction.NONE
        )
        assert jnp.all(
            jnp.isclose(
                mape_elegy(y_true, y_pred), mape_tfk(y_true, y_pred), rtol=0.0001
            )
        )

        ## Prove the loss function
        rng = jax.random.PRNGKey(42)
        y_true = jax.random.randint(rng, shape=(2, 3), minval=0, maxval=2)
        y_pred = jax.random.uniform(rng, shape=(2, 3))
        y_true = y_true.astype(y_pred.dtype)
        loss = elegy.losses.mean_absolute_percentage_error(y_true, y_pred)
        assert loss.shape == (2,)
        assert jnp.array_equal(
            loss,
            100
            * jnp.mean(
                jnp.abs(
                    (y_pred - y_true) / jnp.maximum(jnp.abs(y_true), types.EPSILON)
                ),
                axis=-1,
            ),
        )



from elegy import utils

import jax, jax.numpy as jnp
import numpy as np
from unittest import TestCase
import tempfile, os, pickle
import PIL, urllib

import elegy


class ResNetTest(TestCase):
    def test_basic_predict(self):
        # FIXME: test succeeds if run alone or if run on the cpu-only version of jax
        # test fails with "DNN library is not found" if run on gpu with all other tests together

        model = elegy.Model(elegy.nets.resnet.ResNet18(), run_eagerly=True)
        assert isinstance(model.module, elegy.Module)

        x = np.random.random((2, 224, 224, 3)).astype(np.float32)

        model.init(x)
        y = model.predict(x)

        # update_modules results in a call to `set_default_parameters` for elegy Modules
        # it might be better to have the user call this explicitly to avoid potential OOM
        model.update_modules()

        assert jnp.all(y.shape == (2, 1000))

        # test loading weights from file
        with tempfile.TemporaryDirectory() as tempdir:
            pklpath = os.path.join(tempdir, "delete_me.pkl")
            open(pklpath, "wb").write(
                pickle.dumps(model.module.get_default_parameters())
            )

            new_r18 = elegy.nets.resnet.ResNet18(weights=pklpath)
            y2 = elegy.Model(new_r18, run_eagerly=True).predict(x, initialize=True)

        assert np.allclose(y, y2, rtol=0.001)

    def test_autodownload_pretrained_r18(self):
        fname, _ = urllib.request.urlretrieve(
            "https://upload.wikimedia.org/wikipedia/commons/e/e4/A_French_Bulldog.jpg"
        )
        im = np.array(PIL.Image.open(fname).resize([224, 224])) / np.float32(255)

        r18 = elegy.nets.resnet.ResNet18(weights="imagenet")
        with jax.disable_jit():
            assert (
                elegy.Model(r18).predict(im[np.newaxis], initialize=True).argmax()
                == 245
            )

    def test_autodownload_pretrained_r50(self):
        fname, _ = urllib.request.urlretrieve(
            "https://upload.wikimedia.org/wikipedia/commons/e/e4/A_French_Bulldog.jpg"
        )
        im = np.array(PIL.Image.open(fname).resize([224, 224])) / np.float32(255)

        r50 = elegy.nets.resnet.ResNet50(weights="imagenet")
        with jax.disable_jit():
            assert (
                elegy.Model(r50).predict(im[np.newaxis], initialize=True).argmax()
                == 245
            )



from elegy import types
import typing as tp
import jax
import jax.numpy as jnp
from elegy import utils, types
from elegy.losses.loss import Loss, Reduction


def binary_crossentropy(
    y_true: jnp.ndarray,
    y_pred: jnp.ndarray,
    from_logits: bool = False,
    label_smoothing: float = 0,
) -> jnp.ndarray:
    assert abs(y_pred.ndim - y_true.ndim) <= 1

    y_true, y_pred = utils.maybe_expand_dims(y_true, y_pred)

    if label_smoothing:
        y_true = y_true * (1.0 - label_smoothing) + 0.5 * label_smoothing

    if from_logits:
        return -jnp.mean(y_true * y_pred - jnp.logaddexp(0.0, y_pred), axis=-1)

    y_pred = jnp.clip(y_pred, types.EPSILON, 1.0 - types.EPSILON)
    return -jnp.mean(
        y_true * jnp.log(y_pred) + (1 - y_true) * jnp.log(1 - y_pred), axis=-1
    )


class BinaryCrossentropy(Loss):
    """
    Computes the cross-entropy loss between true labels and predicted labels.
    Use this cross-entropy loss when there are only two label classes (assumed to
    be 0 and 1). For each example, there should be a single floating-point value
    per prediction.
    In the snippet below, each of the four examples has only a single
    floating-pointing value, and both `y_pred` and `y_true` have the shape
    `[batch_size]`.

    Usage:
    ```python
    y_true = jnp.array([[0., 1.], [0., 0.]])
    y_pred = jnp.array[[0.6, 0.4], [0.4, 0.6]])

    # Using 'auto'/'sum_over_batch_size' reduction type.
    bce = elegy.losses.BinaryCrossentropy()
    result = bce(y_true, y_pred)
    assert jnp.isclose(result, 0.815, rtol=0.01)

    # Calling with 'sample_weight'.
    bce = elegy.losses.BinaryCrossentropy()
    result = bce(y_true, y_pred, sample_weight=jnp.array([1, 0]))
    assert jnp.isclose(result, 0.458, rtol=0.01)

    # Using 'sum' reduction type.
    bce = elegy.losses.BinaryCrossentropy(reduction=elegy.losses.Reduction.SUM)
    result = bce(y_true, y_pred)
    assert jnp.isclose(result, 1.630, rtol=0.01)

    # Using 'none' reduction type.
    bce = elegy.losses.BinaryCrossentropy(reduction=elegy.losses.Reduction.NONE)
    result = bce(y_true, y_pred)
    assert jnp.all(jnp.isclose(result, [0.916, 0.713], rtol=0.01))
    ```


    Usage with the `Elegy` API:
    ```python
    model = elegy.Model(
        module_fn,
        loss=elegy.losses.BinaryCrossentropy(),
        metrics=elegy.metrics.Accuracy(),
        optimizer=optax.adam(1e-3),
    )
    ```
    """

    def __init__(
        self,
        from_logits: bool = False,
        label_smoothing: float = 0,
        reduction: tp.Optional[Reduction] = None,
        weight: tp.Optional[float] = None,
        on: tp.Optional[types.IndexLike] = None,
        **kwargs
    ):
        """
        Initializes `CategoricalCrossentropy` instance.

        Arguments:
            from_logits: Whether `y_pred` is expected to be a logits tensor. By
                default, we assume that `y_pred` encodes a probability distribution.
                **Note - Using from_logits=True is more numerically stable.**
            label_smoothing: Float in [0, 1]. When > 0, label values are smoothed,
                meaning the confidence on label values are relaxed. e.g.
                `label_smoothing=0.2` means that we will use a value of `0.1` for label
                `0` and `0.9` for label `1`"
            reduction: (Optional) Type of `elegy.losses.Reduction` to apply to
                loss. Default value is `SUM_OVER_BATCH_SIZE`. Indicates that the reduction
                option will be determined by the usage context. For almost all cases
                this defaults to `SUM_OVER_BATCH_SIZE`. When used with
                `tf.distribute.Strategy`, outside of built-in training loops such as
                `elegy` `compile` and `fit`, ` or `SUM_OVER_BATCH_SIZE`
                will raise an error.
            weight: Optional weight contribution for the total loss. Defaults to `1`.
            on: A string or integer, or iterable of string or integers, that
                indicate how to index/filter the `y_true` and `y_pred`
                arguments before passing them to `call`. For example if `on = "a"` then
                `y_true = y_true["a"]`. If `on` is an iterable
                the structures will be indexed iteratively, for example if `on = ["a", 0, "b"]`
                then `y_true = y_true["a"][0]["b"]`, same for `y_pred`. For more information
                check out [Keras-like behavior](https://poets-ai.github.io/elegy/guides/modules-losses-metrics/#keras-like-behavior).
        """
        super().__init__(reduction=reduction, weight=weight, on=on, **kwargs)
        self._from_logits = from_logits
        self._label_smoothing = label_smoothing

    def call(
        self,
        y_true: jnp.ndarray,
        y_pred: jnp.ndarray,
        sample_weight: tp.Optional[jnp.ndarray] = None,
    ) -> jnp.ndarray:
        """
        Invokes the `BinaryCrossentropy` instance.

        Arguments:
            y_true: Ground truth values.
            y_pred: The predicted values.
            sample_weight: Acts as a
                coefficient for the loss. If a scalar is provided, then the loss is
                simply scaled by the given value. If `sample_weight` is a tensor of size
                `[batch_size]`, then the total loss for each sample of the batch is
                rescaled by the corresponding element in the `sample_weight` vector. If
                the shape of `sample_weight` is `[batch_size, d0, .. dN-1]` (or can be
                broadcasted to this shape), then each loss element of `y_pred` is scaled
                by the corresponding value of `sample_weight`. (Note on`dN-1`: all loss
                functions reduce by 1 dimension, usually axis=-1.)
        Returns:
            Loss values per sample.
        """

        return binary_crossentropy(
            y_true,
            y_pred,
            from_logits=self._from_logits,
            label_smoothing=self._label_smoothing,
        )



import unittest

import elegy
import jax
import jax.numpy as jnp
from elegy.generalized_module.generalized_module import generalize
from flax import linen


class TestLinenModule(unittest.TestCase):
    def test_basic(self):
        class M(linen.Module):
            @linen.compact
            def __call__(self, x):

                initialized = self.has_variable("batch_stats", "n")

                vn = self.variable("batch_stats", "n", lambda: 0)

                w = self.param("w", lambda key: 2.0)

                if initialized:
                    vn.value += 1

                return x * w

        gm = generalize(M())
        rng = elegy.RNGSeq(42)

        y_true, params, states = gm.init(rng)(x=3.0, y=1)

        assert y_true == 6
        assert params["w"] == 2
        assert states["batch_stats"]["n"] == 0

        params = params.copy(dict(w=10.0))
        y_true, params, states = gm.apply(params, states, training=True, rng=rng)(
            x=3.0, y=1
        )

        assert y_true == 30
        assert params["w"] == 10
        assert states["batch_stats"]["n"] == 1

    def test_summaries(self):
        class ModuleC(linen.Module):
            @linen.compact
            @elegy.flax_summarize
            def __call__(self, x):
                c1 = self.param("c1", lambda _: jnp.ones([5]))
                c2 = self.variable("states", "c2", lambda: jnp.ones([6]))

                x = jax.nn.relu(x)
                elegy.flax_summary(self, "relu", jax.nn.relu, x)

                return x

        class ModuleB(linen.Module):
            @linen.compact
            @elegy.flax_summarize
            def __call__(self, x):
                b1 = self.param("b1", lambda _: jnp.ones([3]))
                b2 = self.variable("states", "b2", lambda: jnp.ones([4]))

                x = ModuleC()(x)

                x = jax.nn.relu(x)
                elegy.flax_summary(self, "relu", jax.nn.relu, x)

                return x

        class ModuleA(linen.Module):
            @linen.compact
            @elegy.flax_summarize
            def __call__(self, x):
                a1 = self.param("a1", lambda _: jnp.ones([1]))
                a2 = self.variable("states", "a2", lambda: jnp.ones([2]))

                x = ModuleB()(x)

                x = jax.nn.relu(x)
                elegy.flax_summary(self, "relu", jax.nn.relu, x)

                return x

        model = elegy.Model(ModuleA())
        model.init(x=jnp.ones([10, 2]))

        summary_text = model.summary(x=jnp.ones([10, 2]), depth=1, return_repr=True)
        assert summary_text is not None

        lines = summary_text.split("\n")

        assert "ModuleB_0" in lines[7]
        assert "ModuleB" in lines[7]
        assert "(10, 2)" in lines[7]
        assert "8" in lines[7]
        assert "32 B" in lines[7]
        assert "10" in lines[7]
        assert "40 B" in lines[7]

        assert "relu" in lines[9]
        assert "(10, 2)" in lines[9]

        assert "*" in lines[11]
        assert "ModuleA" in lines[11]
        assert "(10, 2)" in lines[11]
        assert "1" in lines[11]
        assert "4 B" in lines[11]
        assert "2" in lines[11]
        assert "8 B" in lines[11]

        assert "9" in lines[13]
        assert "36 B" in lines[13]

        assert "12" in lines[13]
        assert "48 B" in lines[13]

        assert "21" in lines[16]
        assert "84 B" in lines[16]



import typing as tp
from abc import ABC, abstractmethod

from elegy import types, utils

REGISTRY: tp.Dict[tp.Type, tp.Type["GeneralizedOptimizer"]] = {}


class ModuleExists(Exception):
    pass


class GeneralizedOptimizer(ABC):
    @abstractmethod
    def __init__(self, optimizer: tp.Any):
        ...

    @abstractmethod
    def init(
        self, rng: types.RNGSeq, net_params: types.NetParams
    ) -> types.OptimizerStates:
        ...

    @abstractmethod
    def apply(
        self,
        net_params: types.NetParams,
        grads: types.Grads,
        optimizer_states: types.OptimizerStates,
        rng: types.RNGSeq,
    ) -> tp.Tuple[types.NetParams, types.OptimizerStates]:
        ...


def register_optimizer_for(
    module_type,
) -> tp.Callable[[tp.Type[GeneralizedOptimizer]], tp.Any]:
    def _lambda(generalized_module_type) -> tp.Any:
        if module_type in REGISTRY:
            raise ModuleExists(
                f"Type {module_type} already registered with {REGISTRY[module_type]}"
            )

        REGISTRY[module_type] = generalized_module_type

        return generalized_module_type

    return _lambda


def generalize_optimizer(optimizer: tp.Any) -> GeneralizedOptimizer:

    if isinstance(optimizer, GeneralizedOptimizer):
        return optimizer

    generalized_module_type: tp.Optional[tp.Type[GeneralizedOptimizer]] = None

    for module_type in REGISTRY:
        if isinstance(optimizer, module_type):
            generalized_module_type = REGISTRY[module_type]

    if generalized_module_type is None:
        raise ValueError(f"No GeneralizedOptimizer found for {optimizer}.")

    return generalized_module_type(optimizer)



import typing as tp

import haiku as hk
import jax
import jax.numpy as jnp

from elegy import utils
from elegy.losses.loss import Loss, Reduction
from elegy import types


class GlobalL1L2(Loss):
    r"""
    A regularizer that applies both L1 and L2 regularization penalties.

    The L1 regularization penalty is computed as:

    $$
    \ell_1\,\,penalty =\ell_1\sum_{i=0}^n|x_i|
    $$

    The L2 regularization penalty is computed as

    $$\ell_2\,\,penalty =\ell_2\sum_{i=0}^nx_i^2$$


    Usage:

    ```python
    model = elegy.Model(
        module_fn,
        loss=[
            elegy.losses.SparseCategoricalCrossentropy(),
            elegy.regularizers.GlobalL1L2(l1=1e-5, l2=1e-4),
        ],
        metrics=lambda: elegy.metrics.SparseCategoricalAccuracy(),
    )
    ```

    Attributes:
        l1: L1 regularization factor.
        l2: L2 regularization factor.
    """

    def __init__(
        self,
        l1=0.0,
        l2=0.0,
        reduction: tp.Optional[Reduction] = None,
        weight: tp.Optional[float] = None,
        **kwargs
    ):  # pylint: disable=redefined-outer-name
        super().__init__(reduction=reduction, weight=weight, **kwargs)

        self.l1 = l1
        self.l2 = l2

    def call(self, states: types.States) -> jnp.ndarray:
        """
        Computes the L1 and L2 regularization penalty simultaneously.

        Arguments:
            net_params: A structure with all the parameters of the model.
        """
        net_params = states.net_params
        regularization: jnp.ndarray = jnp.array(0.0)

        if not self.l1 and not self.l2:
            return regularization

        if self.l1:
            regularization += self.l1 * sum(
                jnp.sum(jnp.abs(p)) for p in jax.tree_leaves(net_params)
            )

        if self.l2:
            regularization += self.l2 * sum(
                jnp.sum(jnp.square(p)) for p in jax.tree_leaves(net_params)
            )

        return regularization



from unittest import TestCase

import jax.numpy as jnp
import tensorflow.keras as tfk
import numpy as np

import elegy


class RecallTest(TestCase):
    def test_basic(self):

        y_true = (np.random.uniform(0, 1, size=(5, 6, 7)) > 0.5).astype(np.float32)
        y_pred = np.random.uniform(0, 1, size=(5, 6, 7))
        sample_weight = np.expand_dims(
            (np.random.uniform(0, 1, size=(6, 7)) > 0.5).astype(int), axis=0
        )

        assert np.allclose(
            tfk.metrics.Recall()(y_true, y_pred),
            elegy.metrics.Recall().call_with_defaults()(
                jnp.asarray(y_true), jnp.asarray(y_pred)
            ),
        )

        assert np.allclose(
            tfk.metrics.Recall(thresholds=0.3)(y_true, y_pred),
            elegy.metrics.Recall(threshold=0.3).call_with_defaults()(
                jnp.asarray(y_true), jnp.asarray(y_pred)
            ),
        )

        assert np.allclose(
            tfk.metrics.Recall(thresholds=0.3)(
                y_true, y_pred, sample_weight=sample_weight
            ),
            elegy.metrics.Recall(threshold=0.3).call_with_defaults()(
                jnp.asarray(y_true),
                jnp.asarray(y_pred),
                sample_weight=jnp.asarray(sample_weight),
            ),
        )

        float_sample_weight = np.random.uniform(0, 1, size=(6, 7))[np.newaxis]
        assert np.allclose(
            tfk.metrics.Recall(thresholds=0.3)(
                y_true, y_pred, sample_weight=float_sample_weight
            ),
            elegy.metrics.Recall(threshold=0.3).call_with_defaults()(
                y_true, y_pred, sample_weight=float_sample_weight
            ),
        )

    def test_cumulative(self):
        tm = tfk.metrics.Recall(thresholds=0.3)
        em = elegy.metrics.Recall(threshold=0.3)

        # 1st run
        y_true = (np.random.uniform(0, 1, size=(5, 6, 7)) > 0.5).astype(np.float32)
        y_pred = np.random.uniform(0, 1, size=(5, 6, 7))
        sample_weight = np.expand_dims(
            (np.random.uniform(0, 1, size=(6, 7)) > 0.5).astype(int), axis=0
        )

        assert np.allclose(
            tm(y_true, y_pred, sample_weight=sample_weight),
            em.call_with_defaults()(
                jnp.asarray(y_true),
                jnp.asarray(y_pred),
                sample_weight=jnp.asarray(sample_weight),
            ),
        )

        # 2nd run
        y_true = (np.random.uniform(0, 1, size=(5, 6, 7)) > 0.5).astype(np.float32)
        y_pred = np.random.uniform(0, 1, size=(5, 6, 7))
        sample_weight = np.expand_dims(
            (np.random.uniform(0, 1, size=(6, 7)) > 0.5).astype(int), axis=0
        )

        assert np.allclose(
            tm(y_true, y_pred, sample_weight=sample_weight),
            em.call_with_defaults()(
                jnp.asarray(y_true),
                jnp.asarray(y_pred),
                sample_weight=jnp.asarray(sample_weight),
            ),
        )



import functools
import typing as tp

from elegy import hooks, types, utils

from .generalized_module import GeneralizedModule, register_module_for

try:
    import flax.linen as nn
    from flax import linen
    from flax.core import FrozenDict, freeze, unfreeze
except ImportError:
    raise types.DependencyUnavailable("Flax is not available")


@register_module_for(nn.Module)
class LinenModule(GeneralizedModule):
    def __init__(self, module: nn.Module):
        self.module = module

    def init(self, rng: types.RNGSeq) -> tp.Callable[..., types.OutputStates]:
        def _lambda(*args, **kwargs):
            def init_fn(*args, **kwargs):
                return self.module.init_with_output(rng.next(), *args, **kwargs)

            y_pred, variables = utils.inject_dependencies(
                init_fn,
                signature_f=self.module.__call__,
            )(
                *args,
                **kwargs,
            )
            assert isinstance(variables, FrozenDict)

            net_states, net_params = (
                variables.pop("params")
                if "params" in variables
                else (variables, FrozenDict())
            )

            return types.OutputStates(y_pred, net_params, net_states)

        return _lambda

    def apply(
        self,
        params: tp.Any,
        states: tp.Any,
        training: bool,
        rng: types.RNGSeq,
    ) -> tp.Callable[..., types.OutputStates]:
        if params is None:
            params = FrozenDict()

        if states is None:
            states = FrozenDict()

        def _lambda(*args, **kwargs):
            def apply_fn(*args, **kwargs):
                variables = dict(params=params, **states)
                return self.module.apply(
                    variables,
                    *args,
                    rngs={"params": rng.next()},
                    mutable=True,
                    **kwargs,
                )

            y_pred, variables = utils.inject_dependencies(
                apply_fn,
                signature_f=self.module.__call__,
            )(
                *args,
                **kwargs,
            )

            net_states, net_params = (
                variables.pop("params")
                if "params" in variables
                else (variables, FrozenDict())
            )

            return types.OutputStates(y_pred, net_params, net_states)

        return _lambda

    def get_summary_params(
        self,
        path: types.Path,
        module: tp.Any,
        value: tp.Any,
        net_params: types.NetParams,
        net_states: types.NetStates,
    ) -> tp.Tuple[tp.Optional[types.Pytree], tp.Optional[types.Pytree]]:

        if net_params is None:
            params_tree = None
        else:
            params_tree = utils.get_path_params(path, net_params)
            # filter out submodules
            if params_tree is not None:
                assert isinstance(module, linen.Module)
                params_tree = {
                    name: value
                    for name, value in params_tree.items()
                    if not name[0].isupper()
                }

        if net_states is None:
            states_tree = None
        else:
            states_tree = {
                collection: utils.get_path_params(path, states)
                for collection, states in net_states.items()
            }
            # filter out submodules
            states_tree = {
                collection: {
                    name: value
                    for name, value in states.items()
                    if assert_id(isinstance(module, linen.Module))
                    and not name[0].isupper()
                }
                for collection, states in states_tree.items()
                if states is not None
            }

        return params_tree, states_tree


def flax_summarize(f):
    @functools.wraps(f)
    def wrapper(self: linen.Module, *args, **kwargs):

        outputs = f(self, *args, **kwargs)

        if hooks.summaries_active():
            path = self.scope.path
            hooks.add_summary(path, self, outputs)

        return outputs

    return wrapper


def flax_summary(
    flax_module: linen.Module,
    name: str,
    f: tp.Any,
    value: types.Scalar,
):
    if hooks.summaries_active():
        path = flax_module.scope.path + (name,)
        hooks.add_summary(path, f, value)


def assert_id(value):
    assert value
    return value



from elegy.types import DependencyUnavailable
from .generalized_module import GeneralizedModule
from . import elegy_module

try:
    from . import linen_module
except DependencyUnavailable as e:
    pass



import logging
from unittest import TestCase


import numpy as np
import optax
import jax

import elegy

np.random.seed(42)


class EarlyStoppingTest(TestCase):
    def test_example(self):
        class MLP(elegy.Module):
            def call(self, x):
                x = elegy.nn.Linear(10)(x)
                x = jax.lax.stop_gradient(x)
                return x

        callback = elegy.callbacks.EarlyStopping(monitor="loss", patience=3)
        # This callback will stop the training when there is no improvement in
        # the for three consecutive epochs.
        model = elegy.Model(
            module=MLP(),
            loss=elegy.losses.MeanSquaredError(),
            optimizer=optax.rmsprop(0.01),
        )
        history = model.fit(
            x=np.ones((5, 20)),
            y=np.zeros((5, 10)),
            epochs=10,
            batch_size=1,
            callbacks=[callback],
            verbose=0,
        )
        assert len(history.history["loss"]) == 4  # Only 4 epochs are run.



# Implementation based on tf.keras.callbacks.py
# https://github.com/tensorflow/tensorflow/blob/v2.2.0/tensorflow/python/keras/callbacks.py
import logging

import numpy as np

from .callback import Callback


class ModelCheckpoint(Callback):
    """
    Callback to save the Elegy model or model weights at some frequency.

    `ModelCheckpoint` callback is used in conjunction with training using
    `model.fit()` to save a model or weights at some
    interval, so the model or weights can be loaded later to continue the training
    from the state saved.

    A few options this callback provides include:

    - Whether to only keep the model that has achieved the "best performance" so
        far, or whether to save the model at the end of every epoch regardless of
        performance.
    - Definition of 'best'; which quantity to monitor and whether it should be
        maximized or minimized.
    - The frequency it should save at. Currently, the callback supports saving at
        the end of every epoch, or after a fixed number of training batches.

    Example:

    ```python
    EPOCHS = 10
    checkpoint_path = '/tmp/checkpoint'
    model_checkpoint_callback = elegy.callbacks.ModelCheckpoint(
        path=checkpoint_path,
        monitor='val_acc',
        mode='max',
        save_best_only=True)

    # Model is saved at the end of every epoch, if it's the best seen
    # so far.
    model.fit(epochs=EPOCHS, callbacks=[model_checkpoint_callback])

    # The model status (that are considered the best) are loaded into the model.
    model.load(checkpoint_path)
    ```

    """

    def __init__(
        self,
        path: str,
        monitor: str = "val_loss",
        verbose: int = 0,
        save_best_only: bool = False,
        mode: str = "auto",
        save_freq: str = "epoch",
        period: int = 1,
    ):
        """
        Arguments:
            path: string, path to directory to save the model state. `path` can contain
                named formatting options, which will be filled the value of `epoch` and
                keys in `logs` (passed in `on_epoch_end`). For example: if `path` is
                `weights.{epoch:02d}-{val_loss:.2f}`, then the model checkpoints
                will be saved with the epoch number and the validation loss in the
                filename.
            monitor: quantity to monitor.
            verbose: verbosity mode, 0 or 1.
            save_best_only: if `save_best_only=True`, the latest best model according
                to the quantity monitored will not be overwritten.
                If `path` doesn't contain formatting options like `{epoch}` then
                `path` will be overwritten by each new better model.
            mode: one of {auto, min, max}. If `save_best_only=True`, the decision to
                overwrite the current save file is made based on either the maximization
                or the minimization of the monitored quantity. For `val_acc`, this
                should be `max`, for `val_loss` this should be `min`, etc. In `auto`
                mode, the direction is automatically inferred from the name of the
                monitored quantity.
            save_freq: `'epoch'` or integer. When using `'epoch'`, the callback saves
                the model after each epoch. When using integer, the callback saves the
                model at end of this many batches. Note that if the saving isn't aligned
                to epochs, the monitored metric may potentially be less reliable (it
                could reflect as little as 1 batch, since the metrics get reset every
                epoch). Defaults to `'epoch'`
            period: the number of epochs between which the model is saved. This only works
                if `save_freq` is 'epoch', otherwise the `save_freq` will override
                this period.
        """
        super(ModelCheckpoint, self).__init__()
        self.monitor = monitor
        self.verbose = verbose
        self.path = path
        self.save_best_only = save_best_only
        self.save_freq = save_freq
        self.period = period
        self.epochs_since_last_save = 0
        self._batches_seen_since_last_saving = 0

        if mode not in ["auto", "min", "max"]:
            logging.warning(
                "ModelCheckpoint mode %s is unknown, " "fallback to auto mode.", mode
            )
            mode = "auto"

        if mode == "min":
            self.monitor_op = np.less
            self.best = np.Inf
        elif mode == "max":
            self.monitor_op = np.greater
            self.best = -np.Inf
        else:
            if "acc" in self.monitor or self.monitor.startswith("fmeasure"):
                self.monitor_op = np.greater
                self.best = -np.Inf
            else:
                self.monitor_op = np.less
                self.best = np.Inf

        if self.save_freq != "epoch" and not isinstance(self.save_freq, int):
            raise ValueError("Unrecognized save_freq: {}".format(self.save_freq))

    def set_model(self, model):
        self.model = model

    def on_batch_end(self, batch, logs=None):
        if self._implements_train_batch_hooks():
            logs = logs or {}
            self._batches_seen_since_last_saving += 1
            if self._batches_seen_since_last_saving >= self.save_freq:
                self._save_model(epoch=self._current_epoch, logs=logs)
                self._batches_seen_since_last_saving = 0

    def on_epoch_begin(self, epoch, logs=None):
        self._current_epoch = epoch

    def on_epoch_end(self, epoch, logs=None):
        self.epochs_since_last_save += 1
        # pylint: disable=protected-access
        if self.save_freq == "epoch":
            self._save_model(epoch=epoch, logs=logs)

    def _save_model(self, epoch, logs):
        """Saves the model.

        Arguments:
            epoch: the epoch this iteration is in.
            logs: the `logs` dict passed in to `on_batch_end` or `on_epoch_end`.
        """
        logs = logs or {}

        if (
            isinstance(self.save_freq, int)
            or self.epochs_since_last_save >= self.period
        ):
            self.epochs_since_last_save = 0
            path = self._get_file_path(epoch, logs)

            # try:
            if self.save_best_only:
                current = logs.get(self.monitor)
                if current is None:
                    logging.warning(
                        "Can save best model only with %s available, " "skipping.",
                        self.monitor,
                    )
                else:
                    if self.monitor_op(current, self.best):
                        if self.verbose > 0:
                            print(
                                "\nEpoch %05d: %s improved from %0.5f to %0.5f,"
                                " saving model to %s"
                                % (
                                    epoch + 1,
                                    self.monitor,
                                    self.best,
                                    current,
                                    path,
                                )
                            )
                        self.best = current
                        self.model.save(path)
                    else:
                        if self.verbose > 0:
                            print(
                                "\nEpoch %05d: %s did not improve from %0.5f"
                                % (epoch + 1, self.monitor, self.best)
                            )
            else:
                if self.verbose > 0:
                    print("\nEpoch %05d: saving model to %s" % (epoch + 1, path))

                self.model.save(path)

            # except IOError as e:
            #     # `e.errno` appears to be `None` so checking the content of `e.args[0]`.
            #     if "is a directory" in six.ensure_str(e.args[0]):
            #         raise IOError(
            #             "Please specify a non-directory path for "
            #             "ModelCheckpoint. Filepath used is an existing "
            #             "directory: {}".format(path)
            #         )

    def _get_file_path(self, epoch, logs):
        """Returns the file path for checkpoint."""
        # pylint: disable=protected-access
        try:
            # `path` may contain placeholders such as `{epoch:02d}` and
            # `{mape:.2f}`. A mismatch between logged metrics and the path's
            # placeholders can cause formatting to fail.
            return self.path.format(epoch=epoch + 1, **logs)
        except KeyError as e:
            raise KeyError(
                'Failed to format this callback path: "{}". '
                "Reason: {}".format(self.path, e)
            )

    def _implements_train_batch_hooks(self):
        # If save_freq="epoch", batch-level hooks don't need to be run.
        return isinstance(self.save_freq, int)



import typing as tp
import unittest

import elegy
import jax.numpy as jnp
import numpy as np


class ModelCoreTest(unittest.TestCase):
    def test_init(self):
        N = 0

        class Model(elegy.model.model_core.ModelCore):
            def states_step(self):
                return elegy.States(
                    net_params=None,
                    net_states=None,
                    metrics_states=None,
                    optimizer_states=None,
                )

            def init_step(self, states):
                states = states.maybe_update(**self.states_step())
                _, states = self.train_step(states)

                return states

            def pred_step(self, states):
                nonlocal N
                N = N + 1

                return elegy.PredStep(
                    y_pred=None,
                    states=states.update(net_params=1, net_states=2),
                )

            def test_step(self, states):
                _, states = self.pred_step(states)
                return elegy.TestStep(0, {}, states.update(metrics_states=3))

            def train_step(self, states):
                _, logs, states = self.test_step(states)
                return elegy.TrainStep(logs, states.update(optimizer_states=4))

        model = Model()

        assert N == 0
        assert not hasattr(model.states, "net_params")
        assert not hasattr(model.states, "net_states")
        assert not hasattr(model.states, "metrics_states")
        assert not hasattr(model.states, "optimizer_states")

        model.init_on_batch()

        assert N == 1
        assert model.states.net_params == 1
        assert model.states.net_states == 2
        assert model.states.metrics_states == 3
        assert model.states.optimizer_states == 4

        model.init_on_batch()
        assert N == 1

    def test_pred_step(self):
        class Model(elegy.model.model_core.ModelCore):
            def init_step(self, x, states):
                _, states = self.pred_step(x, states, True)
                return states

            def pred_step(self, x, states, initializing):
                if initializing:
                    states = elegy.States(net_states=0)
                else:
                    states = elegy.States(net_states=states.net_states + 1)

                return elegy.PredStep(
                    y_pred=1,
                    states=states,
                )

        model = Model()
        assert not hasattr(model.states, "net_states")

        model.init_on_batch(x=(np.array(1.0)))
        preds = model.predict_on_batch(x=(np.array(1.0)))
        assert preds == 1
        assert model.states.net_states == 1
        assert not hasattr(model.states, "net_params")

        model.run_eagerly = False

        preds = model.predict_on_batch(x=(np.array(1.0)))
        assert preds == 1
        assert model.states.net_states == 2

    def test_test_step(self):
        class Model(elegy.model.model_core.ModelCore):
            def init_step(self, states):
                _, _, states = self.test_step(states, True)
                return states

            def test_step(self, states, initializing):
                return elegy.TestStep(
                    loss=0.1,
                    logs=dict(loss=1.0),
                    states=elegy.States(metrics_states=0)
                    if initializing
                    else elegy.States(metrics_states=states.metrics_states + 1),
                )

            def train_step(self):
                ...

        model = Model()
        assert not hasattr(model.states, "metrics_states")

        logs = model.test_on_batch(x=(np.array(1.0)), y=(1.0,))
        assert logs["loss"] == 1
        assert model.states.metrics_states == 1
        assert not hasattr(model.states, "net_params")

        model.run_eagerly = False

        logs = model.test_on_batch(x=(np.array(1.0)), y=(1.0,))
        assert logs["loss"] == 1
        assert model.states.metrics_states == 2

    def test_train_step(self):
        class Model(elegy.model.model_core.ModelCore):
            def init_step(self, states):
                _, states = self.train_step(states, True)
                return states

            def train_step(self, states, initializing):
                return elegy.TrainStep(
                    logs=dict(loss=2.0),
                    states=elegy.States(optimizer_states=0)
                    if initializing
                    else elegy.States(optimizer_states=states.optimizer_states + 1),
                )

        model = Model()
        assert not hasattr(model.states, "optimizer_states")

        logs = model.train_on_batch(x=(np.array(1.0)), y=(1.0,))
        assert logs["loss"] == 2
        assert model.states.optimizer_states == 1
        assert not hasattr(model.states, "net_params")

        model.run_eagerly = False

        logs = model.train_on_batch(x=(np.array(1.0)), y=(1.0,))
        assert logs["loss"] == 2
        assert model.states.optimizer_states == 2
